"""Render results in a terminal"""

from collections import (
    defaultdict,
)
import string
from datalad.ui import ui

NA_STRING = 'N/A'  # we might want to make it configurable via config


class nagen(object):
    """A helper to provide a desired missing value if no value is known

    Usecases

    - could be used as a generator for `defaultdict`
    - since it returns itself upon getitem, should work even for complex
      nested dictionaries/lists .format templates
    """
    def __init__(self, missing=NA_STRING):
        self.missing = missing

    def __repr__(self):
        cls = self.__class__.__name__
        args = str(self.missing) if self.missing != NA_STRING else ''
        return '%s(%s)' % (cls, args)

    def __str__(self):
        return self.missing

    def __getitem__(self, *args):
        return self

    def __getattr__(self, item):
        return self


def nadict(*items):
    """A generator of default dictionary with the default nagen"""
    dd = defaultdict(nagen)
    dd.update(*items)
    return dd


class DefaultOutputFormatter(string.Formatter):
    """A custom formatter for default output rendering using .format
    """
    # TODO: make missing configurable?
    def __init__(self, missing=nagen()):
        """
        Parameters
        ----------
        missing: string, optional
          What to output for the missing values
        """
        super(DefaultOutputFormatter, self).__init__()
        self.missing = missing

    def _d(self, msg, *args):
        # print("   HERE %s" % (msg % args))
        pass

    def get_value(self, key, args, kwds):
        assert not args
        self._d("get_value: %r %r %r", key, args, kwds)
        return kwds.get(key, self.missing)

    # def get_field(self, field_name, args, kwds):
    #     assert not args
    #     self._d("get_field: %r args=%r kwds=%r" % (field_name, args, kwds))
    #     try:
    #         out = string.Formatter.get_field(self, field_name, args, kwds)
    #     except Exception as exc:
    #         # TODO needs more than just a value
    #         return "!ERR %s" % exc


class DefaultOutputRenderer(object):
    """A default renderer for .format'ed output line
    """
    def __init__(self, format):
        self.format = format
        # We still need custom output formatter since at the "first level"
        # within .format template all items there is no `nadict`
        self.formatter = DefaultOutputFormatter()

    @classmethod
    def _dict_to_nadict(cls, v):
        """Traverse datastructure and replace any regular dict with nadict"""
        if isinstance(v, list):
            return [cls._dict_to_nadict(x) for x in v]
        elif isinstance(v, dict):
            return nadict((k, cls._dict_to_nadict(x)) for k, x in v.items())
        else:
            return v

    def __call__(self, x, **kwargs):
        dd = nadict(
            (k, nadict({k_.replace(':', '#'): self._dict_to_nadict(v_)
             for k_, v_ in v.items()})
             if isinstance(v, dict) else v)
            for k, v in x.items()
        )

        msg = self.formatter.format(self.format, **dd)
        return ui.message(msg)



# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Test create publication target on GIN"""


from datalad.api import create_sibling_gin
from datalad.tests.utils_pytest import (
    assert_in_results,
    skip_if_no_network,
    with_tempfile,
)

from .test_create_sibling_ghlike import check4real


def check_push(ds):
    # create a file and push it to GIN to see of the
    # access is set up properly
    (ds.pathobj / 'file').write_text('some')
    ds.save()
    assert_in_results(
        ds.push(to='gin', result_renderer='disabled'),
        action='copy',
        status='ok',
        path=str(ds.pathobj / 'file')
    )


@skip_if_no_network
@with_tempfile
def test_gin(path=None):
    check4real(
        create_sibling_gin,
        path,
        'gin',
        'https://gin.g-node.org',
        # when testing locally, you might want to use your
        # own GIN account to not have to fiddle with the key
        # setup
        #'api/v1/repos/mih/{reponame}',
        'api/v1/repos/dataladtester/{reponame}',
        access_protocol='https-ssh',
        moretests=check_push,
    )



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Obsolete module: moved to `local.run_procedure`
"""

import warnings
warnings.warn(
    "RunProcedure has been moved to datalad.local.run_procedure. "
    "This module was deprecated in 0.16.0, and will be removed in a future "
    "release. Please adjust the import.",
    DeprecationWarning)

# Import command class to ease 3rd-party transitions
from datalad.local.run_procedure import *



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""High-level interface for removing dataset content

"""

__docformat__ = 'restructuredtext'

import logging
from os.path import lexists
import warnings

from datalad.core.local.save import Save
from datalad.core.local.status import get_paths_by_ds
from datalad.distributed.drop import Drop
from datalad.distribution.dataset import (
    datasetmethod,
    require_dataset,
    EnsureDataset,
)
from datalad.interface.base import (
    Interface,
    build_doc,
    eval_results,
)
from datalad.interface.common_opts import (
    jobs_opt,
    save_message_opt,
)
from datalad.support.constraints import (
    EnsureChoice,
    EnsureNone,
    EnsureStr,
)
from datalad.support.gitrepo import GitRepo
from datalad.support.param import Parameter
from datalad.utils import (
    ensure_list,
    rmtree,
)

lgr = logging.getLogger('datalad.local.remove')


@build_doc
class Remove(Interface):
    """Remove components from datasets

    Removing "unlinks" a dataset component, such as a file or subdataset, from
    a dataset. Such a removal advances the state of a dataset, just like adding
    new content. A remove operation can be undone, by restoring a previous
    dataset state, but might require re-obtaining file content and subdatasets
    from remote locations.

    This command relies on the 'drop' command for safe operation. By default,
    only file content from datasets which will be uninstalled as part of
    a removal will be dropped. Otherwise file content is retained, such that
    restoring a previous version also immediately restores file content access,
    just as it is the case for files directly committed to Git. This default
    behavior can be changed to always drop content prior removal, for cases
    where a minimal storage footprint for local datasets installations is
    desirable.

    Removing a dataset component is always a recursive operation. Removing a
    directory, removes all content underneath the directory too. If
    subdatasets are located under a to-be-removed path, they will be
    uninstalled entirely, and all their content dropped. If any subdataset
    can not be uninstalled safely, the remove operation will fail and halt.

    .. versionchanged:: 0.16
       More in-depth and comprehensive safety-checks are now performed by
       default.
       The ``if_dirty||--if-dirty`` argument is ignored, will be removed in
       a future release, and can be removed for a safe-by-default behavior. For
       other cases consider the ``reckless||--reckless`` argument.
       The ``save||--save`` argument is ignored and will be removed in a future
       release, a dataset modification is now always saved. Consider save's
       ``amend||--amend`` argument for post-remove fix-ups.
       The ``recursive||--recursive`` argument is ignored, and will be removed
       in a future release. Removal operations are always recursive, and the
       parameter can be stripped from calls for a safe-by-default behavior.

    .. deprecated:: 0.16
       The ``check||--check`` argument will be removed in a future release.
       It needs to be replaced with ``reckless||--reckless``.
    """
    _params_ = dict(
        dataset=Parameter(
            args=("-d", "--dataset"),
            metavar="DATASET",
            doc="""specify the dataset to perform remove from.
            If no dataset is given, the current working directory is used
            as operation context""",
            constraints=EnsureDataset() | EnsureNone()),
        path=Parameter(
            args=("path",),
            metavar="PATH",
            doc="path of a dataset or dataset component to be removed",
            nargs="*",
            constraints=EnsureStr() | EnsureNone()),
        drop=Parameter(
            args=("--drop",),
            doc="""which dataset components to drop prior removal. This
            parameter is passed on to the underlying drop operation as
            its 'what' argument.""",
            # we must not offer a 'nothing' which would bypass
            # the `drop()` call. The implementation completely
            # relies on `drop()` for all safety measures.
            # instead `drop(reckless=kill)` must be used to fast-kill
            # things
            constraints=EnsureChoice('datasets', 'all')),
        jobs=jobs_opt,
        message=save_message_opt,
        # XXX deprecate!
        save=Parameter(
            args=("--nosave",),
            dest='save',
            action="store_false",
            doc="""DEPRECATED and IGNORED; use `save --amend` instead"""),
        recursive=Parameter(
            args=("--recursive", '-r',),
            action='store_const',
            const=None,
            doc="""DEPRECATED and IGNORED: removal is always a recursive
            operation"""),
    )
    # inherit some from Drop
    # if_dirty and check as deprecated
    for p in ('reckless', 'if_dirty', 'check',):
        _params_[p] = Drop._params_[p]

    _examples_ = [
        dict(text="Permanently remove a subdataset (and all further subdatasets contained in it) from a dataset",
             code_py="remove(dataset='path/to/dataset', path='path/to/subds')",
             code_cmd="datalad remove -d <path/to/dataset> <path/to/subds>"),
             dict(text="Permanently remove a superdataset (with all subdatasets) from the filesystem",
             code_py="remove(dataset='path/to/dataset')",
             code_cmd="datalad remove -d <path/to/dataset>"),
             dict(text="DANGER-ZONE: Fast wipe-out a dataset and all its subdataset, bypassing all safety checks",
             code_py="remove(dataset='path/to/dataset', reckless='kill')",
             code_cmd="datalad remove -d <path/to/dataset> --reckless kill"),
    ]

    @staticmethod
    @datasetmethod(name='remove')
    @eval_results
    def __call__(
            path=None,
            *,
            dataset=None,
            drop='datasets',
            reckless=None,
            message=None,
            jobs=None,
            # deprecated below
            recursive=None,
            check=None,
            save=None,
            if_dirty=None):

        # deprecate checks
        if if_dirty is not None:
            warnings.warn(
                "The `if_dirty` argument of `datalad remove` is ignored, "
                "it can be removed for a safe-by-default behavior. For "
                "other cases consider the `reckless` argument.",
                DeprecationWarning)

        if save is not None:
            warnings.warn(
                "The `save` argument of `datalad remove` is ignored. "
                "A dataset modification is always saved. Consider "
                "`save --amend` if post-remove fix-ups are needed.",
                DeprecationWarning)

        if recursive is not None:
            warnings.warn(
                "The `recursive` argument of `datalad remove` is ignored. "
                "Removal operations are always recursive, and the parameter "
                "can be stripped from calls for a safe-by-default behavior. ",
                DeprecationWarning)

        if check is not None:
            warnings.warn(
                "The `check` argument of `datalad remove` is deprecated, "
                "use the `reckless` argument instead.",
                DeprecationWarning)

        if check is False:
            if reckless is not None:
                raise ValueError(
                    'Must not use deprecated `check` argument, and new '
                    '`reckless` argument together with `datalad remove`.')
            reckless = 'availability'

        refds = require_dataset(dataset, check_installed=True,
                                purpose='remove')
        # same path resolution that drop will do
        paths_by_ds, errors = get_paths_by_ds(
            refds, dataset, ensure_list(path),
            # super-mode will readily tell us which datasets to
            # save as the end
            subdsroot_mode='super')

        drop_success = True
        for res in Drop.__call__(
                dataset=dataset,
                path=path,
                what=drop,
                reckless=reckless,
                recursive=True,
                recursion_limit=None,
                jobs=jobs,
                result_xfm=None,
                return_type='generator',
                result_renderer='disabled',
                # delegate error handling here
                on_failure='ignore'):
            if res.get('status') not in ('ok', 'notneeded'):
                drop_success = False
            yield res

        if not drop_success:
            # there will be 'rm -rf' below, so play safe
            lgr.debug('Observed drop failure, will not attempt remove')
            return

        for dpath, paths in paths_by_ds.items():
            for delpath in ([dpath] if paths is None else paths):
                if lexists(str(delpath)):
                    # here we still have something around on the
                    # filesystem. There is no need to fiddle with
                    # Git, just wipe it out. A later save() will
                    # act on it properly
                    if delpath.is_dir():
                        lgr.debug('Remove directory: %s', delpath)
                        rmtree(delpath)
                    # cannot use .exists() must forsee dead symlinks
                    else:
                        lgr.debug('Remove file: %s', delpath)
                        delpath.unlink()
                    continue
                else:
                    # if we get here, there is nothing on the file system
                    # anymore at this path. Either because the parent
                    # dataset vanished already, or because we dropped a
                    # dataset. `save()` will properly unregistered
                    # from the parents at the end. nothing else to do
                    pass

        if not refds.is_installed():
            # we already dropped the whole thing
            return

        for res in Save.__call__(
                dataset=dataset,
                path=path,
                # we might have removed the reference dataset by now, recheck
                message=message if message else '[DATALAD] removed content',
                return_type='generator',
                result_renderer='disabled',
                result_xfm=None,
                result_filter=None,
                on_failure='ignore'):
            if res.get('action') == 'delete':
                # normalize to previous remove results
                res['action'] = 'remove'
            yield res



from datalad.customremotes.main import main as super_main
from datalad.distributed.ora_remote import ORARemote
from datalad.support.annexrepo import AnnexRepo


class DeprecatedRIARemote(ORARemote):
    """This is a shim for backwards compatibility with the old and archived
    git-annex-ria-remote, which the current ORA remote is based on. Providing
    this remote allows datasets that are configured with the old name (and the
    respective config names) to still work.
    However, this is intended to be somewhat temporary and be replaced by
    another implementation that actually migrates from ria to ora once we
    settled for an approach.
    """

    def __init__(self, annex):
        super().__init__(annex)

    def initremote(self):
        self.message("special remote type 'ria' is deprecated. Consider "
                     "migrating to 'ora'.", type='info')
        super().initremote(self)

    def _load_local_cfg(self):
        """Overwrite _load_local_cfg in order to initialize attributes with
        deprecated 'ria' configs if they exist and then go on to let 'super' do
        it's thing"""
        self._repo = AnnexRepo(self.gitdir)
        self.storage_host = \
            self._repo.config.get(f"annex.ria-remote.{self.name}.ssh-host")
        self.store_base_path = \
            self._repo.config.get(f"annex.ria-remote.{self.name}.base-path")
        self.force_write = \
            self._repo.config.get(f"annex.ria-remote.{self.name}.force-write")
        self.ignore_remote_config = \
            self._repo.config.get(f"annex.ria-remote.{self.name}.ignore-remote-config")
        super()._load_local_cfg()


def main():
    """cmdline entry point"""
    super_main(
        cls=DeprecatedRIARemote,
        remote_name='ria',
        description=\
        "transport file content to and from datasets hosted in RIA stores",
    )



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Plumbing command for dataset installation"""


import logging
from argparse import REMAINDER
from typing import Dict

from datalad.cmd import CommandError
from datalad.config import ConfigManager
from datalad.distribution.dataset import (
    Dataset,
    EnsureDataset,
    datasetmethod,
    require_dataset,
    resolve_path,
)
from datalad.interface.base import (
    Interface,
    build_doc,
    eval_results
)
from datalad.interface.common_opts import (
    location_description,
    reckless_opt,
)
from datalad.interface.results import get_status_dict
from datalad.support.annexrepo import AnnexRepo
from datalad.support.constraints import (
    EnsureKeyChoice,
    EnsureNone,
    EnsureStr,
)
from datalad.support.exceptions import CapturedException
from datalad.support.network import (
    RI,
    PathRI,
)
from datalad.support.param import Parameter
from datalad.utils import (
    knows_annex,
    PurePath,
    rmtree,
)

from .clone_utils import (  # needed because other code imports it from here
    _check_autoenable_special_remotes,
    _format_clone_errors,
    _generate_candidate_clone_sources,
    _get_remote,
    _get_tracking_source,
    _map_urls,
    _test_existing_clone_target,
    _try_clone_candidates,
    decode_source_spec,
    # RIA imports needed b/c datalad-next imports it from here ATM;
    # Remove after core was released and next dropped the ria patch.
    postclone_preannex_cfg_ria,
    postclonecfg_ria,
)

__docformat__ = 'restructuredtext'

lgr = logging.getLogger('datalad.core.distributed.clone')


@build_doc
class Clone(Interface):
    """Obtain a dataset (copy) from a URL or local directory

    The purpose of this command is to obtain a new clone (copy) of a dataset
    and place it into a not-yet-existing or empty directory. As such `clone`
    provides a strict subset of the functionality offered by `install`. Only a
    single dataset can be obtained, and immediate recursive installation of
    subdatasets is not supported. However, once a (super)dataset is installed
    via `clone`, any content, including subdatasets can be obtained by a
    subsequent `get` command.

    Primary differences over a direct `git clone` call are 1) the automatic
    initialization of a dataset annex (pure Git repositories are equally
    supported); 2) automatic registration of the newly obtained dataset as a
    subdataset (submodule), if a parent dataset is specified; 3) support
    for additional resource identifiers (DataLad resource identifiers as used
    on datasets.datalad.org, and RIA store URLs as used for store.datalad.org
    - optionally in specific versions as identified by a branch or a tag; see
    examples); and 4) automatic configurable generation of alternative access
    URL for common cases (such as appending '.git' to the URL in case the
    accessing the base URL failed).

    In case the clone is registered as a subdataset, the original URL passed to
    `clone` is recorded in `.gitmodules` of the parent dataset in addition
    to the resolved URL used internally for git-clone. This allows to preserve
    datalad specific URLs like ria+ssh://... for subsequent calls to `get` if
    the subdataset was locally removed later on.

    || PYTHON >>By default, the command returns a single Dataset instance for
    an installed dataset, regardless of whether it was newly installed ('ok'
    result), or found already installed from the specified source ('notneeded'
    result).<< PYTHON ||

    URL mapping configuration

    'clone' supports the transformation of URLs via (multi-part) substitution
    specifications. A substitution specification is defined as a configuration
    setting 'datalad.clone.url-substition.<seriesID>' with a string containing
    a match and substitution expression, each following Python's regular
    expression syntax. Both expressions are concatenated to a single string
    with an arbitrary delimiter character. The delimiter is defined by
    prefixing the string with the delimiter. Prefix and delimiter are stripped
    from the expressions (Example: ",^http://(.*)$,https://\\1").  This setting
    can be defined multiple times, using the same '<seriesID>'.  Substitutions
    in a series will be applied incrementally, in order of their definition.
    The first substitution in such a series must match, otherwise no further
    substitutions in a series will be considered. However, following the first
    match all further substitutions in a series are processed, regardless
    whether intermediate expressions match or not. Substitution series themselves
    have no particular order, each matching series will result in a candidate
    clone URL. Consequently, the initial match specification in a series should
    be as precise as possible to prevent inflation of candidate URLs.

    .. seealso::

      :ref:`handbook:3-001`
        More information on Remote Indexed Archive (RIA) stores
    """
    # by default ignore everything but install results
    # i.e. no "add to super dataset"
    result_filter = EnsureKeyChoice('action', ('install',))
    # very frequently this command will yield exactly one installed dataset
    # spare people the pain of going through a list by default
    return_type = 'item-or-list'
    # as discussed in #1409 and #1470, we want to return dataset instances
    # matching what is actually available after command completion (and
    # None for any failed dataset installation)
    result_xfm = 'successdatasets-or-none'

    _examples_ = [
        dict(text="Install a dataset from GitHub into the current directory",
             code_py="clone("
             "source='https://github.com/datalad-datasets/longnow"
             "-podcasts.git')",
             code_cmd="datalad clone "
             "https://github.com/datalad-datasets/longnow-podcasts.git"),
        dict(text="Install a dataset into a specific directory",
             code_py="""\
             clone(source='https://github.com/datalad-datasets/longnow-podcasts.git',
                   path='myfavpodcasts')""",
             code_cmd="""\
             datalad clone https://github.com/datalad-datasets/longnow-podcasts.git \\
             myfavpodcasts"""),
        dict(text="Install a dataset as a subdataset into the current dataset",
             code_py="""\
             clone(dataset='.',
                   source='https://github.com/datalad-datasets/longnow-podcasts.git')""",
             code_cmd="datalad clone -d . "
             "https://github.com/datalad-datasets/longnow-podcasts.git"),
        dict(text="Install the main superdataset from datasets.datalad.org",
             code_py="clone(source='///')",
             code_cmd="datalad clone ///"),
        dict(text="Install a dataset identified by a literal alias from store.datalad.org",
             code_py="clone(source='ria+http://store.datalad.org#~hcp-openaccess')",
             code_cmd="datalad clone ria+http://store.datalad.org#~hcp-openaccess"),
        dict(
            text="Install a dataset in a specific version as identified by a "
                 "branch or tag name from store.datalad.org",
            code_py="clone(source='ria+http://store.datalad.org#76b6ca66-36b1-11ea-a2e6-f0d5bf7b5561@myidentifier')",
            code_cmd="datalad clone ria+http://store.datalad.org#76b6ca66-36b1-11ea-a2e6-f0d5bf7b5561@myidentifier"),
        dict(
            text="Install a dataset with group-write access permissions",
            code_py=\
            "clone(source='http://example.com/dataset', reckless='shared-group')",
            code_cmd=\
            "datalad clone http://example.com/dataset --reckless shared-group"),
    ]

    _params_ = dict(
        dataset=Parameter(
            args=("-d", "--dataset"),
            doc="""(parent) dataset to clone into. If given, the newly cloned
            dataset is registered as a subdataset of the parent. Also, if given,
            relative paths are interpreted as being relative to the parent
            dataset, and not relative to the working directory.""",
            constraints=EnsureDataset() | EnsureNone()),
        source=Parameter(
            args=("source",),
            metavar='SOURCE',
            doc="""URL, DataLad resource identifier, local path or instance of
            dataset to be cloned""",
            constraints=EnsureStr()),
        path=Parameter(
            args=("path",),
            metavar='PATH',
            nargs="?",
            doc="""path to clone into.  If no `path` is provided a
            destination path will be derived from a source URL
            similar to :command:`git clone`"""),
        git_clone_opts=Parameter(
            args=("git_clone_opts",),
            metavar='GIT CLONE OPTIONS',
            nargs=REMAINDER,
            doc="""[PY: A list of command line arguments PY][CMD: Options CMD]
            to pass to :command:`git clone`. [CMD: Any argument specified after
            SOURCE and the optional PATH will be passed to git-clone. CMD] Note
            that not all options will lead to viable results. For example
            '--single-branch' will not result in a functional annex repository
            because both a regular branch and the git-annex branch are
            required. Note that a version in a RIA URL takes precedence over
            '--branch'."""),
        description=location_description,
        reckless=reckless_opt,
    )

    @staticmethod
    @datasetmethod(name='clone')
    @eval_results
    def __call__(
            source,
            path=None,
            git_clone_opts=None,
            *,
            dataset=None,
            description=None,
            reckless=None,
        ):
        # did we explicitly get a dataset to install into?
        # if we got a dataset, path will be resolved against it.
        # Otherwise path will be resolved first.
        ds = require_dataset(
            dataset, check_installed=True, purpose='clone') \
            if dataset is not None else dataset
        refds_path = ds.path if ds else None

        # legacy compatibility
        if reckless is True:
            # so that we can forget about how things used to be
            reckless = 'auto'

        if isinstance(source, Dataset):
            source = source.path

        if source == path:
            # even if they turn out to be identical after resolving symlinks
            # and more sophisticated witchcraft, it would still happily say
            # "it appears to be already installed", so we just catch an
            # obviously pointless input combination
            raise ValueError(
                "clone `source` and destination `path` are identical [{}]. "
                "If you are trying to add a subdataset simply use `save`".format(
                    path))

        if path is not None:
            path = resolve_path(path, dataset)

        # derive target from source:
        if path is None:
            # we got nothing but a source. do something similar to git clone
            # and derive the path from the source and continue
            # since this is a relative `path`, resolve it:
            # we are not going to reuse the decoded URL, as this is done for
            # all source candidates in clone_dataset(), we just use to determine
            # a destination path here in order to perform a bunch of additional
            # checks that shall not pollute the helper function
            source_ = decode_source_spec(
                source, cfg=None if ds is None else ds.config)
            path = resolve_path(source_['default_destpath'], dataset)
            lgr.debug("Determined clone target path from source")
        lgr.debug("Resolved clone target path to: '%s'", path)

        # there is no other way -- my intoxicated brain tells me
        assert(path is not None)

        result_props = dict(
            action='install',
            logger=lgr,
            refds=refds_path,
            source_url=source)

        try:
            # this will implicitly cause pathlib to run a bunch of checks
            # whether the present path makes any sense on the platform
            # we are running on -- we don't care if the path actually
            # exists at this point, but we want to abort early if the path
            # spec is determined to be useless
            # we can do strict=False since we are 3.6+
            path.resolve(strict=False)
        except OSError as e:
            ce = CapturedException(e)
            yield get_status_dict(
                status='error',
                path=path,
                message=('cannot handle target path: %s', ce),
                exception=ce,
                **result_props)
            return

        destination_dataset = Dataset(path)
        result_props['ds'] = destination_dataset

        if ds is not None and ds.pathobj not in path.parents:
            yield get_status_dict(
                status='error',
                message=("clone target path '%s' not in specified target dataset '%s'",
                         path, ds),
                **result_props)
            return

        # perform the actual cloning operation
        clone_failure = False
        for r in clone_dataset(
                [source],
                destination_dataset,
                reckless,
                description,
                result_props,
                cfg=None if ds is None else ds.config,
                clone_opts=git_clone_opts,
                ):
            if r['status'] in ['error', 'impossible']:
                clone_failure = True
            yield r

        if clone_failure:
            # do not proceed saving anything if cloning failed
            return

        # TODO handle any 'version' property handling and verification using a
        # dedicated public helper

        if ds is not None:
            # we created a dataset in another dataset
            # -> make submodule
            actually_saved_subds = False
            for r in ds.save(
                    path,
                    # Note, that here we know we don't save anything but a new
                    # subdataset. Hence, don't go with default commit message,
                    # but be more specific.
                    message="[DATALAD] Added subdataset",
                    return_type='generator',
                    result_filter=None,
                    result_xfm=None,
                    result_renderer='disabled',
                    on_failure='ignore'):
                actually_saved_subds = actually_saved_subds or (
                        r['action'] == 'save' and
                        r['type'] == 'dataset' and
                        r['refds'] == ds.path and
                        r['status'] == 'ok')
                yield r

            # Modify .gitmodules to contain originally given url. This is
            # particularly relevant for postclone routines on a later `get`
            # for that subdataset. See gh-5256.

            if isinstance(RI(source), PathRI):
                # ensure posix paths; Windows paths would neither be meaningful
                # as a committed path nor are they currently stored correctly
                # (see gh-7182).
                # Restricted to when 'source' is identified as a path, b/c this
                # wouldn't work with file-URLs (ria or not):
                #
                # PureWindowsPath("file:///C:/somewhere/path").as_posix() ->
                # 'file:/C:/somewhere/path'
                source = PurePath(source).as_posix()
            if actually_saved_subds:
                # New subdataset actually saved. Amend the modification
                # of .gitmodules.
                # Note, that we didn't allow deviating from git's default
                # behavior WRT a submodule's name vs its path when we made this
                # a new subdataset.
                subds_name = path.relative_to(ds.pathobj)
                ds.repo.call_git(
                    ['config',
                     '--file',
                     '.gitmodules',
                     '--replace-all',
                     'submodule.{}.{}'.format(subds_name,
                                              "datalad-url"),
                     source]
                )
                yield from ds.save('.gitmodules',
                                   amend=True, to_git=True,
                                   result_renderer='disabled',
                                   return_type='generator')
            else:
                # We didn't really commit. Just call `subdatasets`
                # in that case to have the modification included in the
                # post-clone state (whatever that may be).
                ds.subdatasets(path, set_property=[("datalad-url", source)])


def clone_dataset(
        srcs,
        destds,
        reckless=None,
        description=None,
        result_props=None,
        cfg=None,
        checkout_gitsha=None,
        clone_opts=None):
    """Internal helper to perform cloning without sanity checks (assumed done)

    This helper does not handle any saving of subdataset modification or adding
    in a superdataset.

    Parameters
    ----------
    srcs : list
      Any suitable clone source specifications (paths, URLs)
    destds : Dataset
      Dataset instance for the clone destination
    reckless : {None, 'auto', 'ephemeral', 'shared-...'}, optional
      Mode switch to put cloned dataset into unsafe/throw-away configurations, i.e.
      sacrifice data safety for performance or resource footprint. When None
      and `cfg` is specified, use the value of `datalad.clone.reckless`.
    description : str, optional
      Location description for the annex of the dataset clone (if there is any).
    result_props : dict, optional
      Default properties for any yielded result, passed on to get_status_dict().
    cfg : ConfigManager, optional
      Configuration for parent dataset. This will be queried instead
      of the global DataLad configuration.
    checkout_gitsha : str, optional
      If given, a specific commit, identified by shasum, will be checked out after
      cloning. A dedicated follow-up fetch will be performed, if the initial clone
      did not obtain the commit object. Should the checkout of the target commit
      cause a detached HEAD, the previously active branch will be reset to the
      target commit.
    clone_opts : list of str, optional
      Options passed to git-clone. Note that for RIA URLs, the version is
      translated to a --branch argument, and that will take precedence over a
      --branch argument included in this value.

    Yields
    ------
    dict
      DataLad result records
    """
    # apply the two in-house patches, do local to avoid circular imports
    from . import (
        clone_ephemeral,
        clone_ria,
    )

    if not result_props:
        # in case the caller had no specific idea on how results should look
        # like, provide sensible defaults
        result_props = dict(
            action='install',
            logger=lgr,
            ds=destds,
        )
    else:
        result_props = result_props.copy()

    candidate_sources = _generate_candidate_clone_sources(
        destds, srcs, cfg)

    # important test!
    # based on this `rmtree` will happen below after failed clone
    dest_path_existed, stop_props = _test_existing_clone_target(
        destds, candidate_sources)
    if stop_props:
        # something happened that indicates we cannot continue
        # yield and return
        result_props.update(stop_props)
        yield get_status_dict(**result_props)
        return

    if reckless is None and cfg:
        # if reckless is not explicitly given, but we operate on a
        # superdataset, query whether it has been instructed to operate
        # in a reckless mode, and inherit it for the coming clone
        reckless = cfg.get('datalad.clone.reckless', None)

    last_candidate, error_msgs, stop_props = _try_clone_candidates(
        destds=destds,
        candidate_sources=candidate_sources,
        clone_opts=clone_opts or [],
        dest_path_existed=dest_path_existed,
    )
    if stop_props:
        # no luck, report and stop
        result_props.update(stop_props)
        yield get_status_dict(**result_props)
        return
    else:
        # we can record the last attempt as the candidate URL that gave
        # a successful clone
        result_props['source'] = last_candidate

    if not destds.is_installed():
        # we do not have a clone, stop, provide aggregate error message
        # covering all attempts
        yield get_status_dict(
            status='error',
            message=_format_clone_errors(
                destds, error_msgs, last_candidate['giturl']),
            **result_props)
        return

    #
    # At minimum all further processing is all candidate for extension
    # patching.  wrap the whole thing in try-except, catch any exceptions
    # report it as an error results `rmtree` any intermediate and return
    #
    try:
        yield from _post_gitclone_processing_(
            destds=destds,
            cfg=cfg,
            gitclonerec=last_candidate,
            reckless=reckless,
            checkout_gitsha=checkout_gitsha,
            description=description,
        )
    except Exception as e:
        ce = CapturedException(e)
        # the rational for turning any exception into an error result is that
        # we are hadly able to distinguish user-error from an other errors
        yield get_status_dict(
            status='error',
            # XXX A test in core insists on the wrong message type to be used
            #error_message=ce.message,
            message=ce.message,
            exception=ce,
            **result_props,
        )
        rmtree(destds.path, children_only=dest_path_existed)
        return

    # yield successful clone of the base dataset now, as any possible
    # subdataset clone down below will not alter the Git-state of the
    # parent
    yield get_status_dict(status='ok', **result_props)


def _post_gitclone_processing_(
        *,
        destds: Dataset,
        cfg: ConfigManager,
        gitclonerec: Dict,
        reckless: None or str,
        checkout_gitsha: None or str,
        description: None or str,
):
    """Perform git-clone post-processing

    This is helper is called immediately after a Git clone was established.

    The properties of that clone are passed via `gitclonerec`.

    Yields
    ------
    DataLad result records
    """
    dest_repo = destds.repo
    remote = _get_remote(dest_repo)

    yield from _post_git_init_processing_(
        destds=destds,
        cfg=cfg,
        gitclonerec=gitclonerec,
        remote=remote,
        reckless=reckless,
    )

    if knows_annex(destds.path):
        # init annex when traces of a remote annex can be detected
        yield from _pre_annex_init_processing_(
            destds=destds,
            cfg=cfg,
            gitclonerec=gitclonerec,
            remote=remote,
            reckless=reckless,
        )
        dest_repo = _annex_init(
            destds=destds,
            cfg=cfg,
            gitclonerec=gitclonerec,
            remote=remote,
            description=description,
        )
        yield from _post_annex_init_processing_(
            destds=destds,
            cfg=cfg,
            gitclonerec=gitclonerec,
            remote=remote,
            reckless=reckless,
        )

    if checkout_gitsha and \
       dest_repo.get_hexsha(
            dest_repo.get_corresponding_branch()) != checkout_gitsha:
        try:
            postclone_checkout_commit(dest_repo, checkout_gitsha,
                                      remote=remote)
        except Exception:
            # We were supposed to clone a particular version but failed to.
            # This is particularly pointless in case of subdatasets and
            # potentially fatal with current implementation of recursion.
            # see gh-5387
            lgr.debug(
                "Failed to checkout %s, removing this clone attempt at %s",
                checkout_gitsha, destds.path)
            raise

    yield from _pre_final_processing_(
        destds=destds,
        cfg=cfg,
        gitclonerec=gitclonerec,
        remote=remote,
        reckless=reckless,
    )


def _post_git_init_processing_(
        *,
        destds: Dataset,
        cfg: ConfigManager,
        gitclonerec: Dict,
        remote: str,
        reckless: None or str,
):
    """Any post-git-init processing that need not be concerned with git-annex
    """
    if not gitclonerec.get("version"):
        postclone_check_head(destds, remote=remote)

    # act on --reckless=shared-...
    # must happen prior git-annex-init, where we can cheaply alter the repo
    # setup through safe re-init'ing
    if reckless and reckless.startswith('shared-'):
        lgr.debug(
            'Reinitializing %s to enable shared access permissions',
            destds)
        destds.repo.call_git(['init', '--shared={}'.format(reckless[7:])])

    # trick to have the function behave like a generator, even if it
    # (currently) doesn't actually yield anything.
    # but a patched version might want to...so for uniformity with
    # _post_annex_init_processing_() let's do this
    if False:
        yield


def _pre_annex_init_processing_(
        *,
        destds: Dataset,
        cfg: ConfigManager,
        gitclonerec: Dict,
        remote: str,
        reckless: None or str,
):
    """Pre-processing a to-be-initialized annex repository"""
    if reckless == 'auto':
        lgr.debug(
            "Instruct annex to hardlink content in %s from local "
            "sources, if possible (reckless)", destds.path)
        destds.config.set(
            'annex.hardlink', 'true', scope='local', reload=True)

    # trick to have the function behave like a generator, even if it
    # (currently) doesn't actually yield anything.
    if False:
        yield


def _annex_init(
        *,
        destds: Dataset,
        cfg: ConfigManager,
        gitclonerec: Dict,
        remote: str,
        description: None or str,
):
    """Initializing an annex repository"""
    lgr.debug("Initializing annex repo at %s", destds.path)
    # Note, that we cannot enforce annex-init via AnnexRepo().
    # If such an instance already exists, its __init__ will not be executed.
    # Therefore do quick test once we have an object and decide whether to call
    # its _init().
    #
    # Additionally, call init if we need to add a description (see #1403),
    # since AnnexRepo.__init__ can only do it with create=True
    repo = AnnexRepo(destds.path, init=True)
    if not repo.is_initialized() or description:
        repo._init(description=description)
    return repo


def _post_annex_init_processing_(
        *,
        destds: Dataset,
        cfg: ConfigManager,
        gitclonerec: Dict,
        remote: str,
        reckless: None or str,
):
    """Post-processing an annex repository"""
    # convenience aliases
    repo = destds.repo
    ds = destds

    if reckless == 'auto' or (reckless and reckless.startswith('shared-')):
        repo.call_annex(['untrust', 'here'])

    _check_autoenable_special_remotes(repo)

    # we have just cloned the repo, so it has a remote `remote`, configure any
    # reachable origin of origins
    yield from configure_origins(ds, ds, remote=remote)


def _pre_final_processing_(
        *,
        destds: Dataset,
        cfg: ConfigManager,
        gitclonerec: Dict,
        remote: str,
        reckless: None or str,
):
    """Any post-processing after Git and git-annex pieces are fully initialized
    """
    if reckless:
        # store the reckless setting in the dataset to make it
        # known to later clones of subdatasets via get()
        destds.config.set(
            'datalad.clone.reckless', reckless,
            scope='local',
            reload=True)
    else:
        # We would still want to reload configuration to ensure that any of the
        # above git invocations could have potentially changed the config
        # TODO: might no longer be necessary if 0.14.0 adds reloading upon
        # non-readonly commands invocation
        destds.config.reload()

    # trick to have the function behave like a generator, even if it
    # (currently) doesn't actually yield anything.
    if False:
        yield


def postclone_checkout_commit(repo, target_commit, remote="origin"):
    """Helper to check out a specific target commit in a fresh clone.

    Will not check (again) if current commit and target commit are already
    the same!
    """
    # record what branch we were on right after the clone
    active_branch = repo.get_active_branch()
    corr_branch = repo.get_corresponding_branch(branch=active_branch)
    was_adjusted = bool(corr_branch)
    repo_orig_branch = corr_branch or active_branch
    # if we are on a branch this hexsha will be the tip of that branch
    repo_orig_hexsha = repo.get_hexsha(repo_orig_branch)
    # make sure we have the desired commit locally
    # expensive and possibly error-prone fetch conditional on cheap
    # local check
    if not repo.commit_exists(target_commit):
        try:
            repo.fetch(remote=remote, refspec=target_commit)
        except CommandError as e:
            CapturedException(e)
            pass
        # instead of inspecting the fetch results for possible ways
        # with which it could failed to produced the desired result
        # let's verify the presence of the commit directly, we are in
        # expensive-land already anyways
        if not repo.commit_exists(target_commit):
            # there is nothing we can do about this
            # MIH thinks that removing the clone is not needed, as a likely
            # next step will have to be a manual recovery intervention
            # and not another blind attempt
            raise ValueError(
                'Target commit %s does not exist in the clone, and '
                'a fetch that commit from remote failed'
                % target_commit[:8])
    # checkout the desired commit
    repo.call_git(['checkout', target_commit])
    # did we detach?
    if repo_orig_branch and not repo.get_active_branch():
        # trace if current state is a predecessor of the branch_hexsha
        lgr.debug(
            "Detached HEAD after resetting worktree of %s "
            "(original branch: %s)", repo, repo_orig_branch)
        if repo.get_merge_base(
                [repo_orig_hexsha, target_commit]) == target_commit:
            # we assume the target_commit to be from the same branch,
            # because it is an ancestor -- update that original branch
            # to point to the target_commit, and update HEAD to point to
            # that location
            lgr.info(
                "Reset branch '%s' to %s (from %s) to "
                "avoid a detached HEAD",
                repo_orig_branch, target_commit[:8], repo_orig_hexsha[:8])
            branch_ref = 'refs/heads/%s' % repo_orig_branch
            repo.update_ref(branch_ref, target_commit)
            repo.update_ref('HEAD', branch_ref, symbolic=True)
            if was_adjusted:
                # Note: The --force is needed because the adjust branch already
                # exists.
                repo.adjust(options=["--unlock", "--force"])
        else:
            lgr.warning(
                "%s has a detached HEAD, because the target commit "
                "%s has no unique ancestor with branch '%s'",
                repo, target_commit[:8], repo_orig_branch)


def postclone_check_head(ds, remote="origin"):
    repo = ds.repo
    if not repo.commit_exists("HEAD"):
        # HEAD points to an unborn branch. A likely cause of this is that the
        # remote's main branch is something other than master but HEAD wasn't
        # adjusted accordingly.
        #
        # Let's choose the most recently updated remote ref (according to
        # commit date). In the case of a submodule, switching to a ref with
        # commits prevents .update_submodule() from failing. It is likely that
        # the ref includes the registered commit, but we don't have the
        # information here to know for sure. If it doesn't, .update_submodule()
        # will check out a detached HEAD.
        remote_branches = (
            b["refname:strip=2"] for b in repo.for_each_ref_(
                fields="refname:strip=2", sort="-committerdate",
                pattern="refs/remotes/" + remote))
        for rbranch in remote_branches:
            if rbranch in [remote + "/git-annex", "HEAD"]:
                continue
            if rbranch.startswith(remote + "/adjusted/"):
                # If necessary for this file system, a downstream
                # git-annex-init call will handle moving into an
                # adjusted state.
                continue
            repo.call_git(["checkout", "-b",
                           rbranch[len(remote) + 1:],  # drop "<remote>/"
                           "--track", rbranch])
            lgr.debug("Checked out local branch from %s", rbranch)
            return
        lgr.warning("Cloned %s but could not find a branch "
                    "with commits", ds.path)


def configure_origins(cfgds, probeds, label=None, remote="origin"):
    """Configure any discoverable local dataset sibling as a remote

    Parameters
    ----------
    cfgds : Dataset
      Dataset to receive the remote configurations
    probeds : Dataset
      Dataset to start looking for `remote` remotes. May be identical with
      `cfgds`.
    label : int, optional
      Each discovered remote will be configured as a remote under the name
      '<remote>-<label>'. If no label is given, '2' will be used by default,
      given that there is typically a remote named `remote` already.
    remote : str, optional
      Name of the default remote on clone.
    """
    if label is None:
        label = 1
    # let's look at the URL for that remote and see if it is a local
    # dataset
    origin_url = probeds.config.get(f'remote.{remote}.url')
    if not origin_url:
        # no remote with default name, nothing to do
        return
    if not cfgds.config.obtain(
            'datalad.install.inherit-local-origin',
            default=True):
        # no inheritance wanted
        return
    if not isinstance(RI(origin_url), PathRI):
        # not local path
        return

    # no need to reconfigure original/direct remote again
    if cfgds != probeds:
        # prevent duplicates
        known_remote_urls = set(
            cfgds.config.get(r + '.url', None)
            for r in cfgds.config.sections()
            if r.startswith('remote.')
        )
        if origin_url not in known_remote_urls:
            yield from cfgds.siblings(
                'configure',
                # no chance for conflict, can only be the second configured
                # remote
                name='{}-{}'.format(remote, label),
                url=origin_url,
                # fetch to get all annex info
                fetch=True,
                result_renderer='disabled',
                on_failure='ignore',
            )
    # and dive deeper
    # given the clone source is a local dataset, we can have a
    # cheap look at it, and configure its own `remote` as a remote
    # (if there is any), and benefit from additional annex availability
    yield from configure_origins(
        cfgds,
        Dataset(probeds.pathobj / origin_url),
        label=label + 1,
        remote=remote)



"""Components to build the parser instance for the CLI

This module must import (and run) really fast for a responsive CLI.
It is unconditionally imported by the main() entrypoint.
"""

# ATTN!
# All top-imports are limited to functionality that is necessary for the
# non-error case of constructing of a single target command parser only.
# For speed reasons, all other imports necessary for special cases,
# like error handling, must be done conditionally in-line.

import argparse
from collections import defaultdict
from functools import partial
import sys


from datalad import __version__

from .common_args import common_args
from datalad.interface.base import (
    is_api_arg,
    get_cmd_doc,
    get_interface_groups,
    load_interface,
)
from datalad.utils import getargspec
from .interface import (
    alter_interface_docs_for_cmdline,
    get_cmd_ex,
    get_cmdline_command_name,
)
from datalad.support.constraints import EnsureChoice
from .helpers import get_commands_from_groups
from .exec import call_from_parser

# special case imports
#  .helpers import add_entrypoints_to_interface_groups
#  .helpers.get_description_with_cmd_summary
#  .helpers.get_commands_from_groups
#  .utils.get_suggestions_msg,
#  .interface._known_extension_commands
#  .interface._deprecated_commands

import logging
lgr = logging.getLogger('datalad.cli.parser')


help_gist = """\
Comprehensive data management solution

DataLad provides a unified data distribution system built on the Git
and Git-annex. DataLad command line tools allow to manipulate (obtain,
create, update, publish, etc.) datasets and provide a comprehensive
toolbox for joint management of data and code. Compared to Git/annex
it primarily extends their functionality to transparently and
simultaneously work with multiple inter-related repositories."""


# TODO:  OPT look into making setup_parser smarter to become faster
# Now it seems to take up to 200ms to do all the parser setup
# even though it might not be necessary to know about all the commands etc.
# I wondered if it could somehow decide on what commands to worry about etc
# by going through sys.args first
def setup_parser(
        cmdlineargs,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        return_subparsers=False,
        completing=False,
        help_ignore_extensions=False):
    """
    The holy grail of establishing CLI for DataLad's Interfaces

    Parameters
    ----------
    cmdlineargs: sys.argv
      Used to make some shortcuts when construction of a full parser can be
      avoided.
    formatter_class:
      Passed to argparse
    return_subparsers: bool, optional
      is used ATM only by BuildManPage in _datalad_build_support
    completing: bool, optional
      Flag to indicate whether the process was invoked by argcomplete
    help_ignore_extensions: bool, optional
      Prevent loading of extension entrypoints when --help is requested.
      This is enabled when building docs to avoid pollution of generated
      manpages with extensions commands (that should appear in their own
      docs, but not in the core datalad package docs)
    """
    lgr.log(5, "Starting to setup_parser")

    # main parser
    parser = ArgumentParserDisableAbbrev(
        fromfile_prefix_chars=None,
        prog='datalad',
        # usage="%(prog)s ...",
        description=help_gist,
        formatter_class=formatter_class,
        add_help=False,
        # TODO: when dropping support for Python 3.8: uncomment below
        # and use parse_known_args instead of _parse_known_args:
        # # set to False so parse_known_args does not add its error handling
        # # Added while RFing from using _parse_known_args to parse_known_args.
        # exit_on_error=False,
    )

    # common options
    parser_add_common_options(parser)

    # get all interface definitions from datalad-core
    interface_groups = get_interface_groups()

    # try to figure out whether the parser construction can be limited to
    # a single (sub)command -- don't even try to do this, when we are in
    # any of the doc-building capacities -- timing is not relevant there
    status, parseinfo = single_subparser_possible(
        cmdlineargs,
        parser,
        completing,
    ) if not return_subparsers else ('allparsers', None)

    command_provider = 'core'

    if status == 'allparsers' and not help_ignore_extensions:
        from .helpers import add_entrypoints_to_interface_groups
        add_entrypoints_to_interface_groups(interface_groups)

    # when completing and we have no incomplete option or parameter
    # we still need to offer all commands for completion
    if (completing and status == 'allknown') or (
            status == 'subcommand' and parseinfo not in
            get_commands_from_groups(interface_groups)):
        # we know the command is not in the core package
        # still a chance it could be in an extension
        command_provider = 'extension'
        # we need the full help, or we have a potential command that
        # lives in an extension, must load all extension, expensive
        from .helpers import add_entrypoints_to_interface_groups
        # need to load all the extensions and try again
        # TODO load extensions one-by-one and stop when a command was found
        add_entrypoints_to_interface_groups(interface_groups)

        if status == 'subcommand':
            known_commands = get_commands_from_groups(interface_groups)
            if parseinfo not in known_commands:
                # certainly not possible to identify a single parser that
                # could be constructed, but we can be helpful
                # will sys.exit() unless we are completing
                try_suggest_extension_with_command(
                    parser, parseinfo, completing, known_commands)
                # in completion mode we can get here, even for a command
                # that does not exist at all!
                command_provider = None

    # TODO check if not needed elsewhere
    if status == 'help' or completing and status in ('allknown', 'unknownopt'):
        # --help specification was delayed since it causes immediate
        # printout of
        # --help output before we setup --help for each command
        parser_add_common_opt(parser, 'help')

    all_parsers = {}  # name: (sub)parser

    if (completing and status == 'allknown') or status \
            in ('allparsers', 'subcommand', 'error'):
        # parseinfo could be None here, when we could not identify
        # a subcommand, but need to locate matching ones for
        # completion
        # create subparser, use module suffix as cmd name
        subparsers = parser.add_subparsers()
        for _, _, _interfaces \
                in sorted(interface_groups, key=lambda x: x[1]):
            for _intfspec in _interfaces:
                cmd_name = get_cmdline_command_name(_intfspec)
                if status == 'subcommand':
                    # in case only a subcommand is desired, we could
                    # skip some processing
                    if command_provider and cmd_name != parseinfo:
                        # a known command, but know what we are looking for
                        continue
                    if command_provider is None and not cmd_name.startswith(
                            parseinfo):
                        # an unknown command, and has no common prefix with
                        # the current command candidate, not even good
                        # for completion
                        continue
                subparser = add_subparser(
                    _intfspec,
                    subparsers,
                    cmd_name,
                    formatter_class,
                    completing=completing,
                )
                if subparser:  # interface can fail to load
                    all_parsers[cmd_name] = subparser

    # "main" parser is under "datalad" name
    all_parsers['datalad'] = parser
    lgr.log(5, "Finished setup_parser")
    if return_subparsers:
        # TODO why not pull the subparsers from the main parser?
        return all_parsers
    else:
        return parser


def setup_parser_for_interface(parser, cls, completing=False):
    # XXX needs safety check for name collisions
    # XXX allow for parser kwargs customization
    # get the signature, order of arguments is taken from it
    ndefaults = 0
    args, varargs, varkw, defaults = getargspec(
        cls.__call__, include_kwonlyargs=True)
    if defaults is not None:
        ndefaults = len(defaults)
    default_offset = ndefaults - len(args)
    prefix_chars = parser.prefix_chars
    for i, arg in enumerate(args):
        if not is_api_arg(arg):
            continue

        param = cls._params_[arg]
        defaults_idx = default_offset + i
        if param.cmd_args == tuple():
            # explicitly provided an empty sequence of argument names
            # this shall not appear in the parser
            continue

        # set up the parameter
        setup_parserarg_for_interface(
            parser, arg, param, defaults_idx, prefix_chars, defaults,
            completing=completing)


def setup_parserarg_for_interface(parser, param_name, param, defaults_idx,
                                  prefix_chars, defaults, completing=False):
    cmd_args = param.cmd_args
    parser_kwargs = param.cmd_kwargs
    has_default = defaults_idx >= 0
    if cmd_args:
        if cmd_args[0][0] in prefix_chars:
            # TODO: All the Parameter(args=...) values in this code
            # base use hyphens, so there is no point in the below
            # conversion. If it looks like no extensions rely on this
            # behavior either, this could be dropped.
            parser_args = [c.replace('_', '-') for c in cmd_args]
        else:
            # Argparse will not convert dashes to underscores for
            # arguments that don't start with a prefix character, so
            # the above substitution must be avoided so that
            # call_from_parser() can find the corresponding parameter.
            parser_args = cmd_args
    elif has_default:
        # Construct the option from the Python parameter name.
        parser_args = ("--{}".format(param_name.replace("_", "-")),)
    else:
        # If args= wasn't given and its a positional argument in the
        # function, add a positional argument to argparse. If `dest` is
        # specified, we need to remove it from the keyword arguments
        # because add_argument() expects it as the first argument. Note
        # that `arg` shouldn't have a dash here, but `metavar` can be
        # used if a dash is preferred for the command-line help.
        parser_args = (parser_kwargs.pop("dest", param_name),)

    if has_default:
        parser_kwargs['default'] = defaults[defaults_idx]
    if param.constraints is not None:
        parser_kwargs['type'] = param.constraints
    if completing:
        help = None
        # if possible, define choices to enable their completion
        if 'choices' not in parser_kwargs and \
                isinstance(param.constraints, EnsureChoice):
            parser_kwargs['choices'] = [
                c for c in param.constraints._allowed if c is not None]
    else:
        help = _amend_param_parser_kwargs_for_help(
            parser_kwargs, param,
            defaults[defaults_idx] if defaults_idx >= 0 else None)
    # create the parameter, using the constraint instance for type
    # conversion
    parser.add_argument(*parser_args, help=help,
                        **parser_kwargs)


def _amend_param_parser_kwargs_for_help(parser_kwargs, param, default=None):
    if 'metavar' not in parser_kwargs and \
            isinstance(param.constraints, EnsureChoice):
        parser_kwargs['metavar'] = \
            '{%s}' % '|'.join(
                # don't use short_description(), because
                # it also needs to give valid output for
                # Python syntax (quotes...), but here we
                # can simplify to shell syntax where everything
                # is a string
                p for p in param.constraints._allowed
                # in the cmdline None pretty much means
                # don't give the options, so listing it
                # doesn't make sense. Moreover, any non-string
                # value cannot be given and very likely only
                # serves a special purpose in the Python API
                # or implementation details
                if isinstance(p, str))
    help = alter_interface_docs_for_cmdline(param._doc)
    if help:
        help = help.rstrip()
        if help[-1] != '.':
            help += '.'
        if param.constraints is not None:
            help += _get_help_for_parameter_constraint(param)
    if default is not None and \
            not parser_kwargs.get('action', '').startswith('store_'):
        # if it is a flag, in commandline it makes little sense to show
        # showing the Default: (likely boolean).
        # See https://github.com/datalad/datalad/issues/3203
        help += " [Default: %r]" % (default,)
    return help


def _get_help_for_parameter_constraint(param):
    # include value constraint description and default
    # into the help string
    cdoc = alter_interface_docs_for_cmdline(
        param.constraints.long_description())
    if cdoc[0] == '(' and cdoc[-1] == ')':
        cdoc = cdoc[1:-1]
    return '  Constraints: %s' % cdoc


def single_subparser_possible(cmdlineargs, parser, completing):
    """Performs early analysis of the cmdline

    Looks at the first unparsed argument and if a known command,
    would return only that one.

    When a plain command invocation with `--version` is detected, it will be
    acted on directly (until sys.exit(0) to avoid wasting time on unnecessary
    further processing.

    Returns
    -------
    {'error', 'allknown', 'help', 'unknownopt', 'subcommand'}, None or str
        Returns a status label and a parameter for this status.
        'error': parsing failed, 'allknown': the parser successfully
        identified all arguments, 'help': a help request option was found,
        'unknownopt': an unknown or incomplete option was found,
        'subcommand': a potential subcommand name was found. For the latter
        two modes the second return value is the option or command name.
        For all other modes the second return value is None.
    """
    # Before doing anything additional and possibly expensive see may be that
    # we have got the command already
    try:
        parsed_args, unparsed_args = parser._parse_known_args(
            cmdlineargs[1:], argparse.Namespace())
        # before anything handle possible datalad --version
        if not unparsed_args and getattr(parsed_args, 'version', None):
            parsed_args.version()  # will exit with 0
        if not (completing or unparsed_args):
            # there was nothing that could be a command
            fail_with_short_help(
                parser,
                msg="too few arguments, "
                    "run with --help or visit https://handbook.datalad.org",
                exit_code=2)
        lgr.debug("Command line args 1st pass for DataLad %s. "
                  "Parsed: %s Unparsed: %s",
                  __version__, parsed_args, unparsed_args)
    except Exception as exc:
        # this did not work out
        from datalad.support.exceptions import CapturedException
        ce = CapturedException(exc)
        lgr.debug("Early parsing failed with %s", ce)
        return 'error', None

    if not unparsed_args:
        # cannot possibly be a subcommand
        return 'allknown', None

    unparsed_arg = unparsed_args[0]

    # First unparsed could be either unknown option to top level "datalad"
    # or a command. Among unknown could be --help/--help-np which would
    # need to be dealt with
    if unparsed_arg in ('--help', '--help-np', '-h'):
        # not need to try to tune things, all these will result in everything
        # to be imported and parsed
        return 'help', None
    elif unparsed_arg.startswith('-'):  # unknown or incomplete option
        if completing:
            return 'unknownopt', unparsed_arg
        # will sys.exit
        fail_with_short_help(parser,
                             msg=f"unrecognized argument {unparsed_arg}",
                             # matches exit code of InsufficientArgumentsError
                             exit_code=2)
    else:  # potential command to handle
        return 'subcommand', unparsed_arg


def try_suggest_extension_with_command(parser, cmd, completing, known_cmds):
    """If completing=False, this function will trigger sys.exit()"""
    # check if might be coming from known extensions
    from .interface import (
        _known_extension_commands,
        _deprecated_commands,
    )
    extension_commands = {
        c: e
        for e, commands in _known_extension_commands.items()
        for c in commands
    }
    hint = None
    if cmd in extension_commands:
        hint = "Command %s is provided by (not installed) extension %s." \
               % (cmd, extension_commands[cmd])
    elif cmd in _deprecated_commands:
        hint_cmd = _deprecated_commands[cmd]
        hint = "Command %r was deprecated" % cmd
        hint += (" in favor of %r command." % hint_cmd) if hint_cmd else '.'
    if not completing:
        fail_with_short_help(
            parser,
            hint=hint,
            provided=cmd,
            known=list(known_cmds.keys()) + list(extension_commands.keys())
        )


def add_subparser(_intfspec, subparsers, cmd_name, formatter_class,
                  completing=False):
    """Given an interface spec, add a subparser to subparsers under cmd_name
    """
    _intf = load_interface(_intfspec)
    if _intf is None:
        # failed to load, error was already logged
        return

    # compose argparse.add_parser() arguments, focused on docs
    parser_args = dict(formatter_class=formatter_class)
    # use class description, if no explicit description is available
    intf_doc = get_cmd_doc(_intf)
    if not completing:
        parser_args['description'] = alter_interface_docs_for_cmdline(
            intf_doc)
        if hasattr(_intf, '_examples_'):
            intf_ex = alter_interface_docs_for_cmdline(get_cmd_ex(_intf))
            parser_args['description'] += intf_ex

    # create the sub-parser
    subparser = subparsers.add_parser(cmd_name, add_help=False, **parser_args)
    # our own custom help for all commands, we must do it here
    # (not in setup_parser_for_interface()) because the top-level parser must
    # not unconditionally have it available initially
    parser_add_common_opt(subparser, 'help')
    # let module configure the parser
    setup_parser_for_interface(subparser, _intf, completing=completing)
    # and we would add custom handler for --version
    parser_add_version_opt(
        subparser, _intf.__module__.split('.', 1)[0], include_name=True)
    # logger for command
    # configure 'run' function for this command
    plumbing_args = dict(
        # this is the key execution handler
        func=partial(call_from_parser, _intf),
        # use the logger of the module that defined the interface
        logger=logging.getLogger(_intf.__module__),
        subparser=subparser)
    if hasattr(_intf, 'result_renderer_cmdline'):
        plumbing_args['result_renderer'] = _intf.result_renderer_cmdline
    subparser.set_defaults(**plumbing_args)
    return subparser


class ArgumentParserDisableAbbrev(argparse.ArgumentParser):
    # Don't accept abbreviations for long options. This kludge was originally
    # added at a time when our minimum required Python version was below 3.5,
    # preventing us from using allow_abbrev=False. Now our minimum Python
    # version is high enough, but we still can't use allow_abbrev=False because
    # it suffers from the problem described in 6b3f2fffe (BF: cmdline: Restore
    # handling of short options, 2018-07-23).
    #
    # Modified from the solution posted at
    # https://bugs.python.org/issue14910#msg204678
    def _get_option_tuples(self, option_string):
        chars = self.prefix_chars
        if option_string[0] in chars and option_string[1] in chars:
            # option_string is a long flag. Disable abbreviation.
            return []
        return super(ArgumentParserDisableAbbrev, self)._get_option_tuples(
            option_string)


def parser_add_common_opt(parser, opt, names=None, **kwargs):
    opt_tmpl = common_args[opt]
    opt_kwargs = opt_tmpl[1].copy()
    opt_kwargs.update(kwargs)
    if names is None:
        parser.add_argument(*opt_tmpl[0], **opt_kwargs)
    else:
        parser.add_argument(*names, **opt_kwargs)


def parser_add_common_options(parser, version=None):
    """Add all options defined in common_args, but excludes 'help'"""
    # populate with standard options
    for arg in common_args:
        if arg == 'help':
            continue
        parser_add_common_opt(parser, arg)
    # special case version arg
    if version is not None:
        import warnings
        warnings.warn("Passing 'version' to parser_add_common_options "
                      "no longer has an effect "
                      "and will be removed in a future release.",
                      DeprecationWarning)
    parser_add_version_opt(parser, 'datalad', include_name=True, delay=True)


def parser_add_version_opt(parser, mod_name, include_name=False, delay=False):
    """Setup --version option

    Parameters
    ----------
    parser:
    mod_name: str, optional
    include_name: bool, optional
    delay: bool, optional
      If set to True, no action is taken immediately, and rather
      we assign the function which would print the version. Necessary for
      early pre-parsing of the cmdline
    """

    def print_version():
        mod = sys.modules.get(mod_name, None)
        version = getattr(mod, '__version__', None)
        if version is None:
            # Let's use the standard Python mechanism if underlying module
            # did not provide __version__
            try:
                import pkg_resources
                version = pkg_resources.get_distribution(mod_name).version
            except Exception:
                version = "unknown"
        if include_name:
            print("%s %s" % (mod_name, version))
        else:
            print(version)
        sys.exit(0)

    class versionAction(argparse.Action):
        def __call__(self, parser, args, values, option_string=None):
            if delay:
                setattr(args, self.dest, print_version)
            else:
                print_version()

    parser.add_argument(
        "--version",
        nargs=0,
        action=versionAction,
        help=(
            "show the program's version"
            if not mod_name
            else "show the module and its version which provides the command")
    )


def fail_with_short_help(parser=None,
                         msg=None,
                         known=None, provided=None,
                         hint=None,
                         exit_code=1,
                         what="command",
                         out=None):
    """Generic helper to fail
    with short help possibly hinting on what was intended if `known`
    were provided
    """
    out = out or sys.stderr
    if msg:
        out.write("error: %s\n" % msg)
    if not known:
        if parser:
            parser_add_common_opt(parser, 'help')
            # just to appear in print_usage also consistent with --help output
            parser.add_argument("command [command-opts]")
            parser.print_usage(file=out)
    else:
        out.write(
            "datalad: Unknown %s %r.  See 'datalad --help'.\n\n"
            % (what, provided,))
        if provided not in known:
            from datalad.utils import get_suggestions_msg
            out.write(get_suggestions_msg(provided, known))
    if hint:
        out.write("Hint: %s\n" % hint)
    raise SystemExit(exit_code)



"""Helpers used in the clone.py patch"""

__docformat__ = 'restructuredtext'

import logging
import re
from os.path import expanduser
from pathlib import Path
from typing import (
    Dict,
    List,
    Tuple,
)
from urllib.parse import unquote as urlunquote

from datalad.cmd import (
    CommandError,
    GitWitlessRunner,
    StdOutCapture,
)
from datalad.config import ConfigManager
from datalad.distributed.ora_remote import (
    LocalIO,
    RIARemoteError,
    SSHRemoteIO,
)
from datalad.distribution.dataset import Dataset
from datalad.distribution.utils import _get_flexible_source_candidates
from datalad.dochelpers import single_or_plural
from datalad.log import log_progress
from datalad.runner.exception import CommandError
from datalad.support.annexrepo import AnnexRepo
from datalad.support.exceptions import (
    CapturedException,
    DownloadError
)
from datalad.support.gitrepo import GitRepo
from datalad.support.network import (
    RI,
    SSHRI,
    URL,
    DataLadRI,
    download_url,
    get_local_file_url,
    is_url,
)
from datalad.support.strings import get_replacement_dict
from datalad.utils import (
    Path,
    PurePosixPath,
    ensure_bool,
    ensure_list,
    make_tempfile,
    rmtree,
)


__docformat__ = 'restructuredtext'

lgr = logging.getLogger('datalad.core.distributed.clone')


def postclone_preannex_cfg_ria(ds, remote="origin"):

    # We need to annex-ignore the remote before annex-init is called on the clone,
    # due to issues 5186 and 5253 (and we would have done it afterwards anyway).
    # annex/objects in RIA stores is special for several reasons.
    # 1. the remote doesn't know about it (no actual local annex for the remote)
    # 2. RIA may use hashdir mixed, copying data to it via git-annex (if cloned
    #    via ssh or local) would make it see a bare repo and establish a
    #    hashdir lower annex object tree.
    # 3. We want the ORA remote to receive all data for the store, so its
    #    objects could be moved into archives (the main point of a RIA store).

    # Note, that this function might need an enhancement as theoretically a RIA
    # store could also hold simple standard annexes w/o an intended ORA remote.
    # This needs the introduction of a new version label in RIA datasets, making
    # the following call conditional.
    ds.config.set(f'remote.{remote}.annex-ignore', 'true', scope='local')


def postclonecfg_ria(ds, props, remote="origin"):
    """Configure a dataset freshly cloned from a RIA store"""
    repo = ds.repo

    def get_uuid_from_store(store_url):
        # First figure whether we cloned via SSH, HTTP or local path and then
        # get that config file the same way:
        config_content = None
        scheme = store_url.split(':', 1)[0]
        if scheme in ['http', 'https']:
            try:
                config_content = download_url(
                    "{}{}config".format(
                        store_url,
                        '/' if not store_url.endswith('/') else ''))
            except DownloadError as e:
                ce = CapturedException(e)
                lgr.debug("Failed to get config file from source:\n%s", ce)
        elif scheme == 'ssh':
            # TODO: switch the following to proper command abstraction:
            # SSHRemoteIO ignores the path part ATM. No remote CWD! (To be
            # changed with command abstractions). So we need to get that part to
            # have a valid path to the remote's config file:
            cfg_path = PurePosixPath(URL(store_url).path) / 'config'
            io = SSHRemoteIO(store_url)
            try:
                config_content = io.read_file(cfg_path)
            except RIARemoteError as e:
                ce = CapturedException(e)
                lgr.debug("Failed to get config file from source: %s", ce)

        elif scheme == 'file':
            # TODO: switch the following to proper command abstraction:
            io = LocalIO()
            cfg_path = Path(URL(store_url).localpath) / 'config'
            try:
                config_content = io.read_file(cfg_path)
            except (RIARemoteError, OSError) as e:
                ce = CapturedException(e)
                lgr.debug("Failed to get config file from source: %s", ce)
        else:
            lgr.debug("Unknown URL-Scheme %s in %s. Can handle SSH, HTTP or "
                      "FILE scheme URLs.", scheme, props['source'])

        # And read it
        uuid = None
        if config_content:
            runner = GitWitlessRunner()
            try:
                # "git config -f -" can read from stdin; this spares us a
                # temp file
                result = runner.run(
                    ['git', 'config', '-f', '-', 'datalad.ora-remote.uuid'],
                    stdin=config_content.encode(encoding='utf-8'),
                    protocol=StdOutCapture
                )
                uuid = result['stdout'].strip()
            except CommandError as e:
                ce = CapturedException(e)
                # doesn't contain what we are looking for
                lgr.debug("Found no UUID for ORA special remote at "
                          "'%s' (%s)", remote, ce)

        return uuid




    # chances are that if this dataset came from a RIA store, its subdatasets
    # may live there too. Place a subdataset source candidate config that makes
    # get probe this RIA store when obtaining subdatasets
    ria_store_url = props['source'].split('#', maxsplit=1)[0]
    ds.config.set(
        # we use the label 'origin' for this candidate in order to not have to
        # generate a complicated name from the actual source specification.
        # we pick a cost of 200 to sort it before datalad's default candidates
        # for non-RIA URLs, because they prioritize hierarchical layouts that
        # cannot be found in a RIA store
        'datalad.get.subdataset-source-candidate-200origin',
        # use the entire original URL, up to the fragment + plus dataset ID
        # placeholder, this should make things work with any store setup we
        # support (paths, ports, ...)
        ria_store_url + '#{id}',
        scope='local')

    # setup publication dependency, if a corresponding special remote exists
    # and was enabled (there could be RIA stores that actually only have repos)
    # make this function be a generator
    ora_remotes = [s for s in ds.siblings('query', result_renderer='disabled')
                   if s.get('annex-externaltype') == 'ora']
    # get full special remotes' config for access to stored URL
    srs = repo.get_special_remotes() \
        if hasattr(repo, 'get_special_remotes') else dict()

    has_only_disabled_ora = \
        not ora_remotes and \
        any(r.get('externaltype') == 'ora' for r in srs.values())

    def match_in_urls(special_remote_cfg, url_to_match):
        # Figure whether either `url` or `push-url` in an ORA remote's config
        # match a given URL (to a RIA store).
        return special_remote_cfg['url'].startswith(url_to_match) or \
               (special_remote_cfg['push-url'].startswith(url_to_match)
                if 'push-url' in special_remote_cfg else False)

    no_enabled_ora_matches_url = \
        all(not match_in_urls(srs[r['annex-uuid']], ria_store_url)
            for r in ora_remotes)

    if has_only_disabled_ora or no_enabled_ora_matches_url:

        # No ORA remote autoenabled, but configuration known about at least one,
        # or enabled ORA remotes seem to not match clone URL.
        # Let's check the remote's config for datalad.ora-remote.uuid as stored
        # by create-sibling-ria and try enabling that one.
        lgr.debug("Found no autoenabled ORA special remote. Trying to look it "
                  "up in source config ...")

        org_uuid = get_uuid_from_store(props['giturl'])

        # Now, enable it. If annex-init didn't fail to enable it as stored, we
        # wouldn't end up here, so enable with store URL as suggested by the URL
        # we cloned from.
        if org_uuid:
            if org_uuid in srs.keys():
                # TODO: - Double-check autoenable value and only do this when
                #         true?
                #       - What if still fails? -> Annex shouldn't change config
                #         in that case

                # we only need the store:
                new_url = props['source'].split('#')[0]
                try:
                    # local config to overwrite committed URL
                    repo.config.set(
                        f"remote.{srs[org_uuid]['name']}.ora-url",
                        new_url, scope='local')
                    repo.enable_remote(srs[org_uuid]['name'])
                    lgr.info("Reconfigured %s for %s",
                             srs[org_uuid]['name'], new_url)
                    # update ora_remotes for considering publication dependency
                    # below
                    ora_remotes = [s for s in
                                   ds.siblings('query',
                                               result_renderer='disabled')
                                   if s.get('annex-externaltype', None) ==
                                   'ora']
                except CommandError as e:
                    ce = CapturedException(e)
                    lgr.debug("Failed to reconfigure ORA special remote: %s", ce)
                    repo.config.unset(f"remote.{srs[org_uuid]['name']}.ora-url",
                                      scope='local')
            else:
                lgr.debug("Unknown ORA special remote uuid at '%s': %s",
                          remote, org_uuid)

    # Set publication dependency for `remote` on the respective ORA remote:
    if ora_remotes:
        url_matching_remotes = [r for r in ora_remotes
                                if srs[r['annex-uuid']]['url'] == ria_store_url]

        if len(url_matching_remotes) == 1:
            # We have exactly one ORA remote with the same store URL we used for
            # cloning (includes previously reconfigured remote).
            # Set publication dependency:
            yield from ds.siblings('configure',
                                   name=remote,
                                   publish_depends=url_matching_remotes[0]['name'],
                                   result_filter=None,
                                   result_renderer='disabled')

        elif not url_matching_remotes:
            # No matches but we have successfully autoenabled ORA remotes. Could
            # be the same store accessed by different method (cloning via HTTP
            # but special remote access via SSH). We can confidently set
            # publication dependency if the store knows the UUID.
            org_uuid = get_uuid_from_store(props['giturl'])
            uuid_matching_remotes = [r for r in ora_remotes
                                     if r['annex-uuid'] == org_uuid]
            if uuid_matching_remotes:
                # Multiple uuid matches are actually possible via same-as.
                # However, in that case we can't decide which one is supposed to
                # be used with publishing to `remote`.
                if len(uuid_matching_remotes) == 1:
                    yield from ds.siblings(
                        'configure',
                        name=remote,
                        publish_depends=uuid_matching_remotes[0]['name'],
                        result_filter=None,
                        result_renderer='disabled')
                else:
                    lgr.warning(
                        "Found multiple matching ORA remotes. Couldn't decide "
                        "which one publishing to '%s' should depend on: %s."
                        " Consider running 'datalad siblings configure -s "
                        "%s --publish-depends ORAREMOTENAME' to set "
                        "publication dependency manually.",
                        remote,
                        [r['name'] for r in uuid_matching_remotes],
                        remote)

        else:
            # We have multiple ORA remotes with the same store URL we cloned
            # from.
            lgr.warning("Found multiple matching ORA remotes. Couldn't decide "
                        "which one publishing to '%s' should depend on: %s."
                        " Consider running 'datalad siblings configure -s "
                        "%s --publish-depends ORAREMOTENAME' to set "
                        "publication dependency manually.",
                        remote,
                        [r['name'] for r in url_matching_remotes],
                        remote)


def _get_url_mappings(cfg):
    cfg_prefix = 'datalad.clone.url-substitute.'
    # figure out which keys we should be looking for
    # in the active config
    subst_keys = set(k for k in cfg.keys() if k.startswith(cfg_prefix))
    # and in the common config specs
    from datalad.interface.common_cfg import definitions
    subst_keys.update(k for k in definitions if k.startswith(cfg_prefix))
    # TODO a potential sorting of substitution series could be implemented
    # here
    return [
        # decode the rule specifications
        get_replacement_dict(
            # one or more could come out
            ensure_list(
                cfg.get(
                    k,
                    # make sure to pull the default from the common config
                    default=cfg.obtain(k),
                    # we specifically support declaration of multiple
                    # settings to build replacement chains
                    get_all=True)))
        for k in subst_keys
    ]


def _get_installationpath_from_url(url):
    """Returns a relative path derived from the trailing end of a URL

    This can be used to determine an installation path of a Dataset
    from a URL, analog to what `git clone` does.
    """
    ri = RI(url)
    if isinstance(ri, (URL, DataLadRI, SSHRI)):  # decode only if URL
        path = ri.path.rstrip('/')
        path = urlunquote(path) if path else ri.hostname
        if '/' in path:
            path = path.split('/')
            if path[-1] == '.git':
                path = path[-2]
            else:
                path = path[-1]
    else:
        path = Path(url).parts[-1]
    if path.endswith('.git'):
        path = path[:-4]
    return path


def decode_source_spec(spec, cfg=None):
    """Decode information from a clone source specification

    Parameters
    ----------
    spec : str
      Any supported clone source specification
    cfg : ConfigManager, optional
      Configuration will be queried from the instance (i.e. from a particular
      dataset). If None is given, the global DataLad configuration will be
      queried.

    Returns
    -------
    dict
      The value of each decoded property is stored under its own key in this
      dict. By default the following keys are return: 'type', a specification
      type label {'giturl', 'dataladri', 'ria'}; 'source' the original
      source specification; 'giturl' a URL for the source that is a suitable
      source argument for git-clone; 'version' a version-identifer, if present
      (None else); 'default_destpath' a relative path that that can be used as
      a clone destination.
    """
    if cfg is None:
        from datalad import cfg
    # standard property dict composition
    props = dict(
        source=spec,
        version=None,
    )

    # Git never gets to see these URLs, so let's manually apply any
    # rewrite configuration Git might know about.
    # Note: We need to rewrite before parsing, otherwise parsing might go wrong.
    # This is particularly true for insteadOf labels replacing even the URL
    # scheme.
    spec = cfg.rewrite_url(spec)
    # common starting point is a RI instance, support for accepting an RI
    # instance is kept for backward-compatibility reasons.
    # this conversion will raise ValueError for any unrecognized RI
    source_ri = RI(spec) if not isinstance(spec, RI) else spec

    # scenario switch, each case must set 'giturl' at the very minimum
    if isinstance(source_ri, DataLadRI):
        # we have got our DataLadRI as the source, so expand it
        props['type'] = 'dataladri'
        props['giturl'] = source_ri.as_git_url()
    elif isinstance(source_ri, URL) and source_ri.scheme.startswith('ria+'):
        # parse a RIA URI
        dsid, version = source_ri.fragment.split('@', maxsplit=1) \
            if '@' in source_ri.fragment else (source_ri.fragment, None)
        uuid_regex = r'[a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12}'
        if re.match(uuid_regex, dsid):
            trace = '{}/{}'.format(dsid[:3], dsid[3:])
            default_destpath = dsid
        elif dsid.startswith('~'):
            trace = 'alias/{}'.format(dsid[1:])
            default_destpath = dsid[1:]
        else:
            raise ValueError(
                'RIA URI not recognized, no valid dataset ID or other supported '
                'scheme: {}'.format(spec))
        # now we cancel the fragment in the original URL, but keep everything else
        # in order to be able to support the various combinations of ports, paths,
        # and everything else
        source_ri.fragment = ''
        # strip the custom protocol and go with standard one
        source_ri.scheme = source_ri.scheme[4:]
        # take any existing path, and add trace to dataset within the store
        source_ri.path = '{urlpath}{urldelim}{trace}'.format(
            urlpath=source_ri.path if source_ri.path else '',
            urldelim='' if not source_ri.path or source_ri.path.endswith('/') else '/',
            trace=trace,
        )
        props.update(
            type='ria',
            giturl=str(source_ri),
            version=version,
            default_destpath=default_destpath,
        )
    else:
        # let's assume that anything else is a URI that Git can handle
        props['type'] = 'giturl'
        # use original input verbatim
        props['giturl'] = spec

    if 'default_destpath' not in props:
        # if we still have no good idea on where a dataset could be cloned to if no
        # path was given, do something similar to git clone and derive the path from
        # the source
        props['default_destpath'] = _get_installationpath_from_url(props['giturl'])

    return props


def _map_urls(cfg, urls):
    mapping_specs = _get_url_mappings(cfg)
    if not mapping_specs:
        return urls

    mapped = []
    # we process the candidate in order to maintain any prioritization
    # encoded in it (e.g. _get_flexible_source_candidates_for_submodule)
    # if we have a matching mapping replace the URL in its position
    for u in urls:
        # we only permit a single match
        # TODO we likely want to RF this to pick the longest match
        mapping_applied = False
        # try one mapping set at a time
        for mapping_spec in mapping_specs:
            # process all substitution patterns in the specification
            # always operate on strings (could be a Path instance too)
            mu = str(u)
            matched = False
            for match_ex, subst_ex in mapping_spec.items():
                if not matched:
                    matched = re.match(match_ex, mu) is not None
                if not matched:
                    break
                # try to map, would return unchanged, if there is no match
                mu = re.sub(match_ex, subst_ex, mu)
            if mu != u:
                lgr.debug("URL substitution: '%s' -> '%s'", u, mu)
                mapped.append(mu)
                # we could consider breaking after the for effective mapping
                # specification. however, that would mean any generic
                # definition of a broadly matching substitution would derail
                # the entroe system. moreover, suddently order would matter
                # substantially
                mapping_applied = True
        if not mapping_applied:
            # none of the mappings matches, go with the original URL
            # (really original, not the stringified one)
            mapped.append(u)
    return mapped


def _get_tracking_source(ds):
    """Returns name and url of a potential configured source
    tracking remote"""
    vcs = ds.repo
    # if we have a remote, let's check the location of that remote
    # for the presence of the desired submodule

    remote_name, tracking_branch = vcs.get_tracking_branch()
    if not remote_name and isinstance(vcs, AnnexRepo):
        # maybe cloned from a source repo that was in adjusted mode
        # https://github.com/datalad/datalad/issues/3969
        remote_name, tracking_branch = vcs.get_tracking_branch(
            corresponding=False)
    # TODO: better default `None`? Check where we might rely on '':
    remote_url = ''
    if remote_name:
        remote_url = vcs.get_remote_url(remote_name, push=False)

    return remote_name, remote_url


def _generate_candidate_clone_sources(
        destds: Dataset,
        srcs: List,
        cfg: ConfigManager or None) -> List:
    """Convert "raw" clone source specs to candidate URLs

    Returns
    -------
    Each item in the list is a dictionary with clone candidate properties.
    At minimum each dictionary contains a 'giturl' property, with a URL
    value suitable for passing to `git-clone`. Other properties are
    provided by `decode_source_spec()` and are documented there.
    """
    # check for configured URL mappings, either in the given config manager
    # or in the one of the destination dataset, which is typically not existent
    # yet and the process config is then used effectively
    srcs = _map_urls(cfg or destds.config, srcs)

    # decode all source candidate specifications
    # use a given config or pass None to make it use the process config
    # manager. Theoretically, we could also do
    # `cfg or destds.config` as done above, but some tests patch
    # the process config manager
    candidate_sources = [decode_source_spec(s, cfg=cfg) for s in srcs]

    # now expand the candidate sources with additional variants of the decoded
    # giturl, while duplicating the other properties in the additional records
    # for simplicity. The hope is to overcome a few corner cases and be more
    # robust than git clone
    return [
        dict(props, giturl=s) for props in candidate_sources
        for s in _get_flexible_source_candidates(props['giturl'])
    ]


def _test_existing_clone_target(
        destds: Dataset,
        candidate_sources: List) -> Tuple:
    """Check if the clone target exists, inspect it, if so

    Returns
    -------
    (bool, dict or None)
      A flag whether the target exists, and either a dict with properties
      of a result that should be yielded before an immediate return, or
      None, if the processing can continue
    """
    # important test! based on this `rmtree` will happen below after
    # failed clone
    dest_path = destds.pathobj
    dest_path_existed = dest_path.exists()
    if dest_path_existed and any(dest_path.iterdir()):
        if destds.is_installed():
            # check if dest was cloned from the given source before
            # this is where we would have installed this from
            # this is where it was actually installed from
            track_name, track_url = _get_tracking_source(destds)
            try:
                # this will get us track_url in system native path conventions,
                # whenever it is a path (and not a URL)
                # this is needed to match it to any potentially incoming local
                # source path in the 'notneeded' test below
                track_path = str(Path(track_url))
            except Exception as e:
                CapturedException(e)
                # this should never happen, because Path() will let any non-path
                # stringification pass through unmodified, but we do not want any
                # potential crash due to pathlib behavior changes
                lgr.debug("Unexpected behavior of pathlib!")
                track_path = None
            for cand in candidate_sources:
                src = cand['giturl']
                if track_url == src \
                        or (not is_url(track_url)
                            and get_local_file_url(
                                track_url, compatibility='git') == src) \
                        or track_path == expanduser(src):
                    return dest_path_existed, dict(
                        status='notneeded',
                        message=("dataset %s was already cloned from '%s'",
                                 destds,
                                 src),
                    )
        # anything else is an error
        return dest_path_existed, dict(
            status='error',
            message='target path already exists and not empty, '
                    'refuse to clone into target path',
        )
    # found no reason to stop, i.e. empty target dir
    return dest_path_existed, None


def _try_clone_candidates(
        *,
        destds: Dataset,
        candidate_sources: List,
        clone_opts: List,
        dest_path_existed: bool) -> Tuple:
    """Iterate over candidate URLs and attempt a clone

    Parameters
    ----------
    destds: Dataset
      The target dataset the clone should materialize at.
    candidate_sources: list
      Each value is a dict with properties, as returned by
      `_generate_candidate_clone_sources()`
    clone_opts: list
      Options to be passed on to `_try_clone_candidate()`
    dest_path_existed: bool
      Flag whether the target path existed before attempting a clone.

    Returns
    -------
    (dict or None, dict, dict or None)
      The candidate record of the last clone attempt,
      a mapping of candidate URLs to potential error messages they yielded,
      and either a dict with properties of a result that should be yielded
      before an immediate return, or None, if the processing can continue
    """
    log_progress(
        lgr.info,
        'cloneds',
        'Attempting a clone into %s', destds.path,
        unit=' candidates',
        label='Cloning',
        total=len(candidate_sources),
    )
    error_msgs = dict()  # accumulate all error messages formatted per each url
    for cand in candidate_sources:
        log_progress(
            lgr.info,
            'cloneds',
            'Attempting to clone from %s to %s', cand['giturl'], destds.path,
            update=1,
            increment=True)

        tried_url, error, fatal = _try_clone_candidate(
            destds=destds,
            cand=cand,
            clone_opts=clone_opts,
        )

        if error is not None:
            lgr.debug("Failed to clone from URL: %s (%s)",
                      tried_url, error)

            error_msgs[tried_url] = error

            # ready playing field for the next attempt
            if destds.pathobj.exists():
                lgr.debug("Wiping out unsuccessful clone attempt at: %s",
                          destds.path)
                # We must not just rmtree since it might be curdir etc
                # we should remove all files/directories under it
                # TODO stringification can be removed once patlib compatible
                # or if PY35 is no longer supported
                rmtree(destds.path, children_only=dest_path_existed)

        if fatal:
            # cancel progress bar
            log_progress(
                lgr.info,
                'cloneds',
                'Completed clone attempts for %s', destds
            )
            return cand, error_msgs, fatal

        if error is None:
            # do not bother with other sources if succeeded
            break

    log_progress(
        lgr.info,
        'cloneds',
        'Completed clone attempts for %s', destds
    )
    return cand, error_msgs, None


def _try_clone_candidate(
        *,
        destds: Dataset,
        cand: Dict,
        clone_opts: List) -> Tuple:
    """Attempt a clone from a single candidate

    destds: Dataset
      The target dataset the clone should materialize at.
    candidate_sources: list
      Each value is a dict with properties, as returned by
      `_generate_candidate_clone_sources()`
    clone_opts: list
      Options to be passed on to `_try_clone_candidate()`

    Returns
    -------
    (str, str or None, dict or None)
      The first item is the effective URL a clone was attempted from.
      The second item is `None` if the clone was successful, or an
      error message, detailing the failure for the specific URL.
      If the third item is not `None`, it must be a result dict that
      should be yielded, and no further clone attempt (even when
      other candidates remain) will be attempted.
    """
    # right now, we only know git-clone based approaches
    return _try_git_clone_candidate(
        destds=destds,
        cand=cand,
        clone_opts=clone_opts,
    )


def _try_git_clone_candidate(
        *,
        destds: Dataset,
        cand: Dict,
        clone_opts: List) -> Tuple:
    """_try_clone_candidate() using `git-clone`

    Parameters and return value behavior is as described in
    `_try_clone_candidate()`.
    """
    if cand.get('version', None):
        opts = clone_opts + ["--branch=" + cand['version']]
    else:
        opts = clone_opts

    try:
        GitRepo.clone(
            path=destds.path,
            url=cand['giturl'],
            clone_options=opts,
            create=True)

    except CommandError as e:
        ce = CapturedException(e)
        e_stderr = e.stderr

        # MIH thinks this should rather use any of ce's message generating
        # methods, but kept it to avoid behavior changes
        error_msg = e

        if e_stderr and 'could not create work tree' in e_stderr.lower():
            # this cannot be fixed by trying another URL
            re_match = re.match(r".*fatal: (.*)$", e_stderr,
                                flags=re.MULTILINE | re.DOTALL)
            # existential failure
            return cand['giturl'], error_msg, dict(
                status='error',
                message=re_match.group(1).strip()
                if re_match else "stderr: " + e_stderr,
            )

        # failure for this URL
        return cand['giturl'], error_msg, None

    # success
    return cand['giturl'], None, None


def _format_clone_errors(
        destds: Dataset,
        error_msgs: List,
        last_clone_url: str) -> Tuple:
    """Format all accumulated clone errors across candidates into one message

    Returns
    -------
    (str, list)
      Message body and string formatting arguments for it.
    """
    if len(error_msgs):
        if all(not e.stdout and not e.stderr for e in error_msgs.values()):
            # there is nothing we can learn from the actual exception,
            # the exit code is uninformative, the command is predictable
            error_msg = "Failed to clone from all attempted sources: %s"
            error_args = list(error_msgs.keys())
        else:
            error_msg = "Failed to clone from any candidate source URL. " \
                        "Encountered errors per each url were:\n- %s"
            error_args = '\n- '.join(
                '{}\n  {}'.format(url, exc.to_str())
                for url, exc in error_msgs.items()
            )
    else:
        # yoh: Not sure if we ever get here but I felt that there could
        #      be a case when this might happen and original error would
        #      not be sufficient to troubleshoot what is going on.
        error_msg = "Awkward error -- we failed to clone properly. " \
                    "Although no errors were encountered, target " \
                    "dataset at %s seems to be not fully installed. " \
                    "The 'succesful' source was: %s"
        error_args = (destds.path, last_clone_url)
    return error_msg, error_args


def _get_remote(repo: GitRepo) -> str:
    """Return the name of the remote of a freshly clones repo

    Raises
    ------
    RuntimeError
      In case there is no remote, which should never happen.
    """
    remotes = repo.get_remotes(with_urls_only=True)
    nremotes = len(remotes)
    if nremotes == 1:
        remote = remotes[0]
        lgr.debug("Determined %s to be remote of %s", remote, repo)
    elif remotes > 1:
        lgr.warning(
            "Fresh clone %s unexpected has multiple remotes: %s. Using %s",
            repo.path, remotes, remotes[0])
        remote = remotes[0]
    else:
        raise RuntimeError("bug: fresh clone has zero remotes")
    return remote


def _check_autoenable_special_remotes(repo: AnnexRepo):
    """Check and report on misconfigured/dysfunctional special remotes
    """
    srs = {True: [], False: []}  # special remotes by "autoenable" key
    remote_uuids = None  # might be necessary to discover known UUIDs

    repo_config = repo.config
    for uuid, config in repo.get_special_remotes().items():
        sr_name = config.get('name', None)
        if sr_name is None:
            lgr.warning(
                'Ignoring special remote %s because it does not have a name. '
                'Known information: %s',
                uuid, config)
            continue
        sr_autoenable = config.get('autoenable', False)
        try:
            sr_autoenable = ensure_bool(sr_autoenable)
        except ValueError as e:
            CapturedException(e)
            lgr.warning(
                'Failed to process "autoenable" value %r for sibling %s in '
                'dataset %s as bool.'
                'You might need to enable it later manually and/or fix it up '
                'to avoid this message in the future.',
                sr_autoenable, sr_name, repo.path)
            continue

        # If it looks like a type=git special remote, make sure we have up to
        # date information. See gh-2897.
        if sr_autoenable and repo_config.get(
                "remote.{}.fetch".format(sr_name)):
            try:
                repo.fetch(remote=sr_name)
            except CommandError as exc:
                ce = CapturedException(exc)
                lgr.warning("Failed to fetch type=git special remote %s: %s",
                            sr_name, ce)

        # determine whether there is a registered remote with matching UUID
        if uuid:
            if remote_uuids is None:
                remote_uuids = {
                    # Check annex-config-uuid first. For sameas annex remotes,
                    # this will point to the UUID for the configuration (i.e.
                    # the key returned by get_special_remotes) rather than the
                    # shared UUID.
                    (repo_config.get('remote.%s.annex-config-uuid' % r) or
                     repo_config.get('remote.%s.annex-uuid' % r))
                    for r in repo.get_remotes()
                }
            if uuid not in remote_uuids:
                srs[sr_autoenable].append(sr_name)

    if srs[True]:
        lgr.debug(
            "configuration for %s %s added because of autoenable,"
            " but no UUIDs for them yet known for dataset %s",
            # since we are only at debug level, we could call things their
            # proper names
            single_or_plural("special remote",
                             "special remotes", len(srs[True]), True),
            ", ".join(srs[True]),
            repo.path
        )

    if srs[False]:
        # if has no auto-enable special remotes
        lgr.info(
            'access to %s %s not auto-enabled, enable with:\n'
            '\t\tdatalad siblings -d "%s" enable -s %s',
            # but since humans might read it, we better confuse them with our
            # own terms!
            single_or_plural("dataset sibling",
                             "dataset siblings", len(srs[False]), True),
            ", ".join(srs[False]),
            repo.path,
            srs[False][0] if len(srs[False]) == 1 else "SIBLING",
        )



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""This is the main() CLI entryproint"""

# It should start-up and run as fast as possible for a responsive CLI.

# Imports are done inline and as late as possible to avoid paying for
# an unconditional commulative overhead that is only actually needed
# in some special cases.

__docformat__ = 'restructuredtext'

import logging

lgr = logging.getLogger('datalad.cli')

lgr.log(5, "Importing cli.main")

import os
import sys

import datalad

from .parser import setup_parser

# TODO cross-check with unconditional imports in .parser
# special case imports
#   from .helpers import _fix_datalad_ri
#   import platform
#   from .helpers import _parse_overrides_from_cmdline
#   from datalad.utils import chpwd
#   from .utils import setup_exceptionhook
#   from datalad.support.exceptions import ...

# unconditional imports, no meaningful functionality without them
#   from .parser import setup_parser


def _on_msys_tainted_paths():
    """This duplicates datalad.utils.on_msys_tainted_paths

    But it does it while minimizing runtime penalties on all irrelevant
    systems.
    """
    if os.environ.get('MSYSTEM', '')[:4] not in ('MSYS', 'MING'):
        return False
    if 'MSYS_NO_PATHCONV' in os.environ:
        return False
    import platform
    if platform.system().lower() != 'windows':
        return False
    return True


def main(args=sys.argv):
    """Main CLI entrypoint"""
    lgr.log(5, "Starting main(%r)", args)
    # record that we came in via the cmdline
    datalad.__api = 'cmdline'
    completing = "_ARGCOMPLETE" in os.environ
    if completing and 'COMP_LINE' in os.environ:
        import shlex

        # TODO support posix=False too?
        args = shlex.split(os.environ['COMP_LINE']) or args

    if _on_msys_tainted_paths():
        # Possibly present DataLadRIs were stripped of a leading /
        from .helpers import _fix_datalad_ri
        args = [_fix_datalad_ri(s) for s in args]

    from datalad.support.entrypoints import load_extensions

    # load extensions requested by configuration
    # analog to what coreapi is doing for a Python session
    # importantly, load them prior to parser construction, such
    # that CLI tuning is also within reach for extensions
    load_extensions()

    # PYTHON_ARGCOMPLETE_OK
    # TODO possibly construct a dedicated parser just for autocompletion
    # rather than lobotomizing the normal one
    parser = setup_parser(args, completing=completing)
    try:
        import argcomplete
        argcomplete.autocomplete(parser)
    except ImportError:
        pass

    # parse cmd args
    lgr.debug("Parsing known args among %r", args)
    cmdlineargs, unparsed_args = parser.parse_known_args(args[1:])
    # did the parser tell us what command to run?
    has_func = hasattr(cmdlineargs, 'func') and cmdlineargs.func is not None
    if unparsed_args:
        if has_func:
            lgr.error('unknown argument%s: %s',
                's' if len(unparsed_args) > 1 else '',
                unparsed_args if len(unparsed_args) > 1 else unparsed_args[0],
            )
            cmdlineargs.subparser.print_usage()
            sys.exit(1)
        else:
            # store all unparsed arguments
            cmdlineargs.datalad_unparsed_args = unparsed_args

    # pull config overrides from cmdline args and put in effect
    if cmdlineargs.cfg_overrides is not None:
        from .helpers import _parse_overrides_from_cmdline
        datalad.cfg.overrides.update(
            _parse_overrides_from_cmdline(cmdlineargs)
        )
        # enable overrides
        datalad.cfg.reload(force=True)
        # try loading extensions again, in case the configuration
        # added new ones to consider
        load_extensions()

    if 'datalad.runtime.librarymode' in datalad.cfg:
        datalad.enable_librarymode()

    if cmdlineargs.change_path is not None:
        from datalad.utils import chpwd
        for path in cmdlineargs.change_path:
            chpwd(path)

    # check argparse could determine what commands needs to be executed
    if not has_func:
        # just let argparser spit out its error, since there is smth wrong
        parser.parse_args(args)
        # if that one didn't puke -- we should
        parser.print_usage()
        lgr.error("Please specify the command")
        # matches exit code for InsufficientArgumentsError
        sys.exit(2)

    _run(cmdlineargs)


def _run(namespace):
    """Execute a CLI operation

    Depending on CLI debugging options the CLI operation is executed
    in a debug harness or an exception handler.

    Parameters
    ----------
    namespace: Namespace
      Object returned by `ArgumentParser.parse_args()` with fully
      populated and validated CLI command and arguments.

    Raises
    ------
    SystemExit
      When the CLI completed without error (exit 0).
    """
    # execute the command, either with a debugger catching
    # a crash, or with a simplistic exception handler.
    # note that result rendering is happening in the
    # execution handler, when the command-generator is unwound
    ret = _run_with_debugger(namespace) \
        if namespace.common_debug or namespace.common_idebug \
        else _run_with_exception_handler(namespace)

    # all good, not strictly needed, but makes internal testing easier
    sys.exit(0)


def _run_with_debugger(cmdlineargs):
    """Execute the command and drop into debugger if it crashes"""
    from .utils import setup_exceptionhook

    # so we could see/stop clearly at the point of failure
    setup_exceptionhook(ipython=cmdlineargs.common_idebug)
    return cmdlineargs.func(cmdlineargs)


def _run_with_exception_handler(cmdlineargs):
    """Execute the command and perform some reporting
    normalization if it crashes, but otherwise just let it go"""
    # otherwise - guard and only log the summary. Postmortem is not
    # as convenient if being caught in this ultimate except
    try:
        return cmdlineargs.func(cmdlineargs)
    # catch BaseException for KeyboardInterrupt
    except BaseException as exc:
        from datalad.support.exceptions import (
            CapturedException,
            CommandError,
            IncompleteResultsError,
            InsufficientArgumentsError,
        )
        ce = CapturedException(exc)
        # we crashed, it has got to be non-zero for starters
        exit_code = 1
        if isinstance(exc, InsufficientArgumentsError):
            # if the func reports inappropriate usage, give help output
            lgr.error('%s (%s)', ce, exc.__class__.__name__)
            cmdlineargs.subparser.print_usage(sys.stderr)
            exit_code = 2
        elif isinstance(exc, IncompleteResultsError):
            # in general we do not want to see the error again, but
            # present in debug output
            lgr.debug('could not perform all requested actions: %s', ce)
        elif isinstance(exc, CommandError):
            exit_code = _communicate_commanderror(exc) or exit_code
        elif isinstance(exc, KeyboardInterrupt):
            from datalad.ui import ui
            ui.error("\nInterrupted by user while doing magic: %s" % ce)
            exit_code = 3
        else:
            # some unforeseen problem
            lgr.error('%s', ce.format_with_cause())
        sys.exit(exit_code)


def _communicate_commanderror(exc):
    """Behave as if the command ran directly"""
    exc_msg = exc.to_str(include_output=False)
    if exc_msg:
        msg = exc_msg.encode() if isinstance(exc_msg, str) else exc_msg
        os.write(2, msg + b"\n")
    # push any captured output to the respective streams
    for out, stream in ((exc.stdout, 1), (exc.stderr, 2)):
        if out:
            os.write(stream,
                     out.encode() if isinstance(out, str) else out)
    # pass on exit code
    return exc.code


lgr.log(5, "Done importing cli.main")



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Support functionality for extension development"""

import logging
from datalad.interface.common_cfg import (
    _NotGiven,
    _ConfigDefinition,
    definitions as _definitions,
)

__all__ = ['register_config', 'has_config']

lgr = logging.getLogger('datalad.support.extensions')


def register_config(
        name,
        title,
        *,
        default=_NotGiven,
        default_fn=_NotGiven,
        description=None,
        # yes, we shadow type, this is OK
        type=_NotGiven,
        # for manual entry
        dialog=None,
        scope=_NotGiven,
    ):
    """Register a configuration item

    This function can be used by DataLad extensions and other client
    code to register configurations items and their documentation with
    DataLad's configuration management. Specifically, these definitions
    will be interpreted by and acted on by the `configuration` command,
    and `ConfigManager.obtain()`.

    At minimum, each item must be given a name, and a title. Optionally, any
    configuration item can be given a default (or a callable to compute a
    default lazily on access), a type-defining/validating callable (i.e.
    `Constraint`), a (longer) description, a dialog type to enable manual
    entry, and a configuration scope to store entered values in.

    Parameters
    ----------
    name: str
      Configuration item name, in most cases starting with the prefix
      'datalad.' followed by at least a section name, and a variable
      name, e.g. 'datalad.section.variable', following Git's syntax for
      configuration items.
    title: str
      The briefest summary of the configuration item's purpose, typically
      written in the style of a headline for a dialog UI, or that of an
      explanatory inline comment just prior the item definitions.
    default: optional
      A default value that is already known at the time of registering the
      configuration items. Can be of any type.
    default_fn: callable, optional
      A callable to compute a default value lazily on access. The can be
      used, if the actual value is not yet known at the time of registering
      the configuration item, or if the default is expensive to compute
      and its evaluation needs to be deferred to prevent slow startup
      (configuration items are typically defined as one of the first things
      on import).
    description: str, optional
      A longer description to accompany the title, possibly with instructions
      on how a sensible value can be determined, or with details on the
      impact of a configuration switch.
    type: callable, optional
      A callable to perform arbitrary type conversion and validation of value
      (or default values). If validation/conversion fails, the callable
      must raise an arbitrary exception. The `str(callable)` is used as
      a type description.
    dialog: {'yesno', 'question'}
      A type of UI dialog to use when manual value entry is attempted
      (only in interactive sessions, and only when no default is defined.
      `title` and `description` will be displayed in this dialog.
    scope: {'override', 'global', 'local', 'branch'}, optional
      If particular code requests the storage of (manually entered) values,
      but defines no configuration scope, this default scope will be used.

    Raises
    ------
    ValueError
      For missing required, or invalid configuration properties.
    """
    kwargs = dict(
        default=default,
        default_fn=default_fn,
        scope=scope,
        type=type
    )
    kwargs = {k: v for k, v in kwargs.items() if v is not _NotGiven}
    if dialog is not None and not title:
        raise ValueError("Configuration dialog must have a title")
    doc_props = dict(title=title)
    if description:
        doc_props['text'] = description
    # dialog is OK to be None, this is not just about UI, even if
    # the key of the internal data structure seems to suggest that.
    # it is also the source for annotating config listings
    kwargs['ui'] = (dialog, doc_props)
    _definitions[name] = _ConfigDefinition(**kwargs)


def has_config(name):
    """Returns whether a configuration item is registered under the given name

    Parameters
    ----------

    name: str
      Configuration item name

    Returns
    -------
    bool
    """
    return name in _definitions



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""CLI entrypoint for special remotes"""

__docformat__ = 'restructuredtext'


import argparse

from datalad.cli.parser import (
    parser_add_common_opt,
    parser_add_version_opt,
)
from datalad.cli.utils import setup_exceptionhook
from datalad.ui import ui

import logging
lgr = logging.getLogger('datalad.customremotes')


def setup_parser(remote_name, description):
    # setup cmdline args parser
    # main parser
    parser = argparse.ArgumentParser(
        description= \
        f"git-annex-remote-{remote_name} is a git-annex custom special " \
        f"remote to {description}",
        epilog='"DataLad\'s git-annex very special remote"',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        add_help=True,
    )
    # common options
    parser_add_common_opt(parser, 'log_level')
    parser_add_version_opt(parser, 'datalad', include_name=True)
    if __debug__:
        parser.add_argument(
            '--dbg', action='store_true', dest='common_debug',
            help="Catch exceptions and fall into debugger upon exception")
    return parser


def _main(args, cls):
    """Unprotected portion"""
    assert(cls is not None)
    from annexremote import Master
    master = Master()
    remote = cls(master)
    master.LinkRemote(remote)
    master.Listen()
    # cleanup
    if hasattr(remote, 'stop'):
        remote.stop()


def main(args=None, cls=None, remote_name=None, description=None):
    import sys
    parser = setup_parser(remote_name, description)
    # parse cmd args
    args = parser.parse_args(args)

    # stdin/stdout will be used for interactions with annex
    ui.set_backend('annex')

    if args.common_debug:
        # So we could see/stop clearly at the point of failure
        setup_exceptionhook()
        _main(args, cls)
    else:
        # Otherwise - guard and only log the summary. Postmortem is not
        # as convenient if being caught in this ultimate except
        try:
            _main(args, cls)
        except Exception as exc:
            lgr.debug('%s (%s) - passing ERROR to git-annex and exiting',
                      str(exc), exc.__class__.__name__)
            # `SpecialRemote` classes are supposed to catch everything and
            # turn it into a `RemoteError` resulting in an ERROR message to
            # annex. If we end up here, something went wrong outside of the
            # `master.Listen()` call in `_main`.
            # In any case, exiting the special remote process should be
            # accompanied by such an ERROR message to annex rather than a log
            # message.
            print("ERROR %s (%s)" % (str(exc), exc.__class__.__name__))
            sys.exit(1)



""""""

__docformat__ = 'restructuredtext'

import logging
from typing import Dict

from datalad.distribution.dataset import Dataset
# For now kept in clone_utils, to avoid circular import (see datalad-next)
from .clone_utils import (
    postclone_preannex_cfg_ria,
    postclonecfg_ria,
)

from . import clone as mod_clone

# we need to preserve the original functions to be able to call them
# in the patch
orig_post_git_init_processing_ = mod_clone._post_git_init_processing_
orig_pre_final_processing_ = mod_clone._pre_final_processing_


lgr = logging.getLogger('datalad.core.distributed.clone')


def _post_git_init_processing_(
    *,
    destds: Dataset,
    gitclonerec: Dict,
    remote: str,
    **kwargs
):
    yield from orig_post_git_init_processing_(
        destds=destds, gitclonerec=gitclonerec, remote=remote,
        **kwargs)

    # In case of RIA stores we need to prepare *before* annex is called at all
    if gitclonerec['type'] == 'ria':
        postclone_preannex_cfg_ria(destds, remote=remote)


def _pre_final_processing_(
        *,
        destds: Dataset,
        gitclonerec: Dict,
        remote: str,
        **kwargs
):
    if gitclonerec['type'] == 'ria':
        yield from postclonecfg_ria(destds, gitclonerec,
                                    remote=remote)

    yield from orig_pre_final_processing_(
        destds=destds, gitclonerec=gitclonerec, remote=remote,
        **kwargs)


def _apply():
    # apply patch in a function, to be able to easily patch it out
    # and turn off the patch
    lgr.debug(
        'Apply RIA patch to clone.py:_post_git_init_processing_')
    mod_clone._post_git_init_processing_ = _post_git_init_processing_
    lgr.debug(
        'Apply RIA patch to clone.py:_pre_final_processing_')
    mod_clone._pre_final_processing_ = _pre_final_processing_


_apply()
